"""
Synchronize or validate a sparse subrepository export. The sparse export
configuration is stored in the `.sparse-export` file in the root of the
repository. If the file does not exist, then there is nothing to update
or verify.
"""

import argparse
import hashlib
import logging
import os
import shutil
import subprocess
import sys

logger = logging.getLogger(__name__)


def setup_argparser(argparser):
  argparser.add_argument(
      "--log-level", default="warning",
      choices=["debug", "info", "warning", "error"])
  argparser.add_argument(
      "command", choices=["update", "verify"],
      help="whether to update or verify the export")
  argparser.add_argument("--force")


def get_argdict(args):
  out = {}
  for key, value in vars(args).items():
    if key.startswith("_"):
      continue
    out[key] = value
  return out


def copy_file(sourcepath, destpath):
  destdir = os.path.dirname(destpath)
  if not os.path.exists(destdir):
    logger.debug("Making directory %s", destdir)
    os.makedirs(destdir)
  logger.debug("Copying %s -> %s", sourcepath, destpath)
  shutil.copy2(sourcepath, destpath)
  return 0


def read_chunks(infile, chunksize=4096):
  while True:
    chunk = infile.read(chunksize)
    if chunk == b'':
      break
    yield chunk


def hash_file(filepath, hashname="sha1", chunksize=4096):
  hashobj = hashlib.new(hashname)
  with open(filepath, "rb") as infile:
    for chunk in read_chunks(infile):
      hashobj.update(chunk)
  return hashobj


def verify_file(sourcepath, destpath):
  if not os.path.exists(destpath):
    logger.error("Missing %s", destpath)
    return 1

  if hash_file(sourcepath).digest() == hash_file(destpath).digest():
    logger.info("File %s up to date", destpath)
    return 0
  logger.error("File %s differs from source", destpath)
  return 1


def inner_main(command, force):
  if command == "update":
    operation = copy_file
  elif command == "verify":
    operation = verify_file
  else:
    raise ValueError("Unknown command {}".format(command))

  repodir = subprocess.check_output(
      ["git", "rev-parse", "--show-toplevel"]).strip().decode("utf-8")
  exportsdir = os.path.join(repodir, "tangent/tooling/sparse-exports")

  if force:
    logger.debug("Forcing export of %s", force)
    dirname = force
  else:
    if os.path.exists(os.path.join(exportsdir, "iamgroot.txt")):
      # This is the monorepo
      return 0

    sparse_sentinel = os.path.join(repodir, ".sparse-export")
    if not os.path.exists(exportsdir) and os.path.exists(sparse_sentinel):
      # This is the public export repo
      return 0

    dirnames = os.listdir(exportsdir)
    if len(dirnames) != 1:
      logger.error(
          "Invalid sparse export, %s contains too many directories", exportsdir)
      return 1
    dirname = dirnames[0]

  export_dir = os.path.join(exportsdir, dirname)
  returncode = 0
  for directory, _dirnames, filenames in os.walk(export_dir):
    reldir = os.path.relpath(directory, export_dir)
    for filename in filenames:
      if reldir == ".":
        relpath_file = filename
      else:
        relpath_file = os.path.join(reldir, filename)

      sourcepath = os.path.join(directory, filename)
      if relpath_file == "sparse-checkout":
        gitdir = subprocess.check_output(
            ["git", "rev-parse", "--git-dir"]).decode("utf-8").strip()
        destpath = os.path.join(gitdir, "info/sparse-checkout")
      else:
        destpath = os.path.join(repodir, relpath_file)

      returncode |= operation(sourcepath, destpath)
  return returncode


def main():
  logging.basicConfig(level=logging.INFO)
  argparser = argparse.ArgumentParser(description=__doc__)
  setup_argparser(argparser)
  try:
    import argcomplete
    argcomplete.autocomplete(argparser)
  except ImportError:
    pass
  args = argparser.parse_args()
  argdict = get_argdict(args)
  log_level = argdict.pop("log_level")
  logging.getLogger().setLevel(getattr(logging, log_level.upper()))
  return inner_main(**argdict)


if __name__ == "__main__":
  sys.exit(main())



# PYTHON_ARGCOMPLETE_OK
"""
Perform variable substitution in vscode configuration files
"""

import argparse
import io
import logging
import os
import re
import sys

logger = logging.getLogger(__name__)


def split_vars(varlist):
  out = {}
  for var in varlist:
    if "=" not in var:
      raise ValueError(
          "Invalid variable {}, should be in the form of <name>=<value>"
          .format(var))
    key, value = var.split("=")
    out[key] = value
  return out


def reverse_map(subtitutions):
  return {value: key for key, value in subtitutions.items()}


def subfile(content, substitutions):
  def subcallback(match):
    varname = match.group(1)
    return substitutions.get(varname, match.group(0))

  return re.sub(r"\$\{([^\}]+)\}", subcallback, content)


def unsubfile(content, substitutions):
  for varname, value in substitutions.items():
    content = content.replace(value, "${" + varname + "}")
  return content


def main():
  logging.basicConfig()

  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument(
      "direction", choices=["sub", "unsub"], default="sub",
      help="whether to substitute values for variables or vice versa")
  argparser.add_argument(
      "--log-level", choices=["debug", "info", "warning", "error"],
      default="info")
  argparser.add_argument(
      "--var", action="append", dest="vars",
      help="variable to substitue in the form of <name>=<value>."
           " Can be repeated")
  argparser.add_argument(
      "--touch", action="append", default=[],
      help="a list of files to touch in addition")
  argparser.add_argument(
      "files", nargs="*", help="filepaths to modify")

  try:
    import argcomplete
    argcomplete.autocomplete(argparser)
  except ImportError:
    pass

  args = argparser.parse_args()
  logging.getLogger().setLevel(getattr(logging, args.log_level.upper()))

  if args.direction == "sub":
    subfn = subfile
  elif args.direction == "unsub":
    subfn = unsubfile
  else:
    raise ValueError("Unexpected direction {}".format(args.direction))

  substitutions = split_vars(args.vars)
  for filepath in args.files:
    with io.open(filepath, "r", encoding="utf-8") as infile:
      original_content = infile.read()
    new_content = subfn(original_content, substitutions)
    if new_content == original_content:
      logger.info("%s is unchanged", filepath)
      continue
    logger.info("rewriting %s", filepath)
    with io.open(filepath, "w", encoding="utf-8") as outfile:
      outfile.write(new_content)

  for filepath in args.touch:
    os.utime(filepath, None)

  return 0


if __name__ == "__main__":
  sys.exit(main())



from __future__ import print_function
from __future__ import unicode_literals

import re
import sys

from cmakelang import common

# NOTE(josh): inspiration and some bits taken from cmakeast_ and
# cmakelistparsing_
#
# .. _cmake_ast: https://github.com/polysquare/cmake-ast
# .. _cmakelist_parsing: https://github.com/ijt/cmakelists_parsing


class TokenType(common.EnumObject):
  _id_map = {}


TokenType.QUOTED_LITERAL = TokenType(0)
TokenType.LEFT_PAREN = TokenType(1)
TokenType.RIGHT_PAREN = TokenType(2)
TokenType.WORD = TokenType(3)
TokenType.NUMBER = TokenType(4)
TokenType.DEREF = TokenType(5)
TokenType.WHITESPACE = TokenType(6)
TokenType.NEWLINE = TokenType(7)
TokenType.COMMENT = TokenType(8)
TokenType.UNQUOTED_LITERAL = TokenType(9)
TokenType.FORMAT_OFF = TokenType(10)
TokenType.FORMAT_ON = TokenType(11)
TokenType.BRACKET_ARGUMENT = TokenType(12)
TokenType.BRACKET_COMMENT = TokenType(13)
TokenType.BYTEORDER_MARK = TokenType(14)
TokenType.ATWORD = TokenType(15)


class SourceLocation(tuple):
  """
  Named tuple of (line, col, offset)
  """

  # def __init__(self, line, col, offset):
  #   super(SourceLocation, self).__init__((line, col, offset))

  @property
  def line(self):
    return self[0]

  @property
  def col(self):
    return self[1]

  @property
  def offset(self):
    return self[2]

  def __repr__(self):
    return '{}:{}'.format(self.line, self.col)


class Token(object):
  """
  Lexical unit of a listfile.
  """

  def __init__(self, tok_type, spelling, index, begin, end):
    self.type = tok_type
    self.spelling = spelling
    self.index = index
    self.begin = begin
    self.end = end

  @property
  def content(self):
    return self.spelling

  # TODO(josh): get rid of this? Is it used or did I accidentally add it when
  # I meant to add get_location()? Or should this be a property?
  def location(self):
    return self.begin

  def get_location(self):
    return self.begin

  def count_newlines(self):
    return self.spelling.count('\n')

  def __repr__(self):
    """A string representation of this token."""
    return ("Token(type={0}, "
            "content={1}, "
            "line={2}, "
            "col={3})").format(self.type.name, repr(self.spelling), *self.begin)


def tokenize(contents):
  """
  Scan a string and return a list of Token objects representing the contents
  of the cmake listfile.
  """

  # https://cmake.org/cmake/help/v3.0/manual/
  #     cmake-language.7.html#grammar-token-unquoted_legacy
  legacy_pattern = "({})+".format(
      "|".join([
          # Make-style variable like $(MAKE)
          r'(\$\([^\$\(\)]+\))',
          # Quoted-substring
          r'("[^"\\]*(?:\\.[^"\\]*)*")',
          # Any element except whitespace or one of '()#"\'
          r'([^\s\(\)#"\\])',
          # Escape sequences
          # https://cmake.org/cmake/help/v3.0/manual/
          #   cmake-language.7.html#grammar-token-escape_sequence
          r'(\\[\(\)#" \\\$@^trn;])'
      ])
  )

  # https://cmake.org/cmake/help/v3.0/manual/
  #     cmake-language.7.html#unquoted-argument
  unquoted_pattern = "({})+".format(
      "|".join([
          # Any element except whitespace or one of '()#"\'
          r'([^\s\(\)#"\\])',
          # Escape sequences
          # https://cmake.org/cmake/help/v3.0/manual/
          #   cmake-language.7.html#grammar-token-escape_sequence
          r'(\\[\(\)#" \\\$@^trn;])'
      ])
  )

  # Regexes are in priority order. Changing the order may alter the
  # behavior of the lexer
  scanner = re.Scanner([
      # double quoted string
      # NOTE(josh): regex borrowed from
      # https://stackoverflow.com/a/37379449/141023
      (r'(?<![^\s\(])"[^"\\]*(?:\\.[^"\\]*)*"(?![^\s\)])',
       lambda s, t: (TokenType.QUOTED_LITERAL, t)),
      # single quoted string
      (r"(?<![^\s\(])'[^'\\]*(?:\\.[^'\\]*)*'(?![^\s\)])",
       lambda s, t: (TokenType.QUOTED_LITERAL, t)),
      # bracket argument
      (r"(?<![^\s\(])\[(=*)\[.*?\]\1\](?![^\s\)])",
       lambda s, t: (TokenType.BRACKET_ARGUMENT, t)),
      (r"(?<![^\s\(])-?[0-9]+(?![^\s\)\(])",
       lambda s, t: (TokenType.NUMBER, t)),
      # Either a valid function name or variable name.
      (r"(?<![^\s\(])[a-zA-z_][a-zA-Z0-9_]*(?![^\s\)\(])",
       lambda s, t: (TokenType.WORD, t)),
      # A configure_file replacement @<word>@
      # Either a valid function name or variable name.
      (r"(?<![^\s\(])@[a-zA-z_][a-zA-Z0-9_]*@(?![^\s\)\(])",
       lambda s, t: (TokenType.ATWORD, t)),
      # A variable dereference ${<word>}
      (r"(?<![^\s\(])\${[a-zA-z_][a-zA-Z0-9_]*}(?![^\s\)])",
       lambda s, t: (TokenType.DEREF, t)),
      # unquoted_legacy
      (legacy_pattern,
       lambda s, t: (TokenType.UNQUOTED_LITERAL, t)),
      # unquoted_element+
      (unquoted_pattern,
       lambda s, t: (TokenType.UNQUOTED_LITERAL, t)),
      (r"\(", lambda s, t: (TokenType.LEFT_PAREN, t)),
      (r"\)", lambda s, t: (TokenType.RIGHT_PAREN, t)),
      # NOTE(josh): bare carriage returns are very unlikely to be used but
      # just for the case of explicitnes, if we ever encounter any we treat
      # it as a newline
      (r"\r?\n", lambda s, t: (TokenType.NEWLINE, t)),
      (r"\r\n?", lambda s, t: (TokenType.NEWLINE, t)),
      # NOTE(josh): don't match '\s' here or we'll miss some newline tokens
      # TODO(josh): should we match unicode whitespace too?
      (r"[ \t\f\v]+", lambda s, t: (TokenType.WHITESPACE, t)),
      (r"#\s*(cmake-format|cmf): off[^\n]*",
       lambda s, t: (TokenType.FORMAT_OFF, t)),
      (r"#\s*(cmake-format|cmf): on[^\n]*",
       lambda s, t: (TokenType.FORMAT_ON, t)),
      # bracket comment
      (r"#\[(=*)\[.*?\]\1\]", lambda s, t: (TokenType.BRACKET_COMMENT, t)),
      # line comment
      (r"#[^\n]*", lambda s, t: (TokenType.COMMENT, t)),
      # Catch-all for literals which are compound statements.
      (r"([^\s\(\)]+|[^\s\(]*[^\)]|[^\(][^\s\)]*)",
       lambda s, t: (TokenType.UNQUOTED_LITERAL, t)),
  ], re.DOTALL)

  tokens_return = []
  if contents.startswith("\ufeff"):
    tokens_return = [
        Token(tok_type=TokenType.BYTEORDER_MARK,
              spelling=contents[0],
              index=-1,
              begin=SourceLocation((0, 0, 0)),
              end=SourceLocation((0, 0, 0)))]
    contents = contents[1:]

  tokens, remainder = scanner.scan(contents)

  # Now add line, column, and serial number to token objects. We get lineno
  # by maintaining a running count of newline characters encountered among
  # tokens so far, and column count by splitting the most recent token on
  # it's right most newline. Note that line and numbers are 1-indexed to match
  # up with editors but column numbers are zero indexed because its fun to be
  # inconsistent.
  lineno = 1
  col = 0
  offset = 0
  for tok_index, (tok_type, spelling) in enumerate(tokens):
    if sys.version_info[0] < 3:
      assert isinstance(spelling, unicode)
    begin = SourceLocation((lineno, col, offset))

    newlines = spelling.count('\n')
    lineno += newlines
    if newlines:
      col = len(spelling.rsplit('\n', 1)[1])
    else:
      col += len(spelling)

    offset += len(bytearray(spelling, 'utf-8'))
    tokens_return.append(Token(tok_type=tok_type,
                               spelling=spelling,
                               index=tok_index,
                               begin=begin,
                               end=SourceLocation((lineno, col, offset))))

  if remainder:
    raise common.UserError(
        "Lexer Error: failed to tokenize input starting at: {}:{} with:\n {}"
        .format(lineno, col, remainder[:-20]))

  return tokens_return


def parse_bracket_argument(text):
  regex = re.compile(r'^\[(=*)\[(.*)\]\1\]$', re.DOTALL)
  match = regex.match(text)
  assert match, "Failed to match bracket argument pattern in {}".format(text)
  return ('[' + match.group(1) + '[',
          match.group(2),
          ']' + match.group(1) + ']')


def parse_bracket_comment(text):
  prefix, content, suffix = parse_bracket_argument(text[1:])
  return ('#' + prefix, content, suffix)


def get_first_non_whitespace_token(tokens):
  """
  Return the first token in the list that is not whitespace, or None
  """
  for token in tokens:
    if token.type not in (TokenType.WHITESPACE, TokenType.NEWLINE):
      return token
  return None



"""
Ensure that the precommit repository is up-to-date with the released versions
of cmakelang.
"""

from __future__ import absolute_import, unicode_literals

import argparse
import io
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile

from packaging import version as versiontool

logger = logging.getLogger(__name__)


def get_env(deploykey_path):
  env = os.environ.copy()
  if deploykey_path:
    if not os.path.exists(deploykey_path):
      raise RuntimeError(
          "Supplied deploy key {} does not exist".format(deploykey_path))
    env["GIT_SSH_COMMAND"] = (
        "ssh -o IdentitiesOnly=yes -i {} -F /dev/null".format(deploykey_path))
  return env


def get_upstream_versions(remote, deploykey_path):
  versions = []
  regex = re.compile(r"^(\w{40})\s*refs/tags/v(.*)$")

  tag_pairs = subprocess.check_output(
      ["git", "ls-remote", "--tags", remote], env=get_env(deploykey_path)
  ).decode("utf-8").split("\n")

  for line in tag_pairs:
    match = regex.match(line.strip())
    if not match:
      continue

    try:
      sha1 = match.group(1)
      version = versiontool.parse(match.group(2))
    except versiontool.InvalidVersion:
      continue

    versions.append((version, sha1))
  return sorted(versions)


SETUP_PY = """\
from setuptools import setup
setup(
    name="cmakelang-precommit-dummy",
    version="0.0.0",
    install_requires=["cmakelang=={}"]
)
"""


def inner_main(workdir, versionstr, deploykey_path, dry_run, keep):
  version = versiontool.parse(versionstr)
  remote = "git@github.com:cheshirekow/cmake-format-precommit"

  upstream_versions = get_upstream_versions(remote, deploykey_path)
  if not upstream_versions:
    logging.error("Received empty list of upstream versions")
    return 1

  upstream_numbers = [upversion for upversion, _ in upstream_versions]
  if version in upstream_numbers:
    logger.info(
        "Desired version tag %s already exists on the origin,"
        " nothing to do", version)
    return 0

  latest_upstream_version = upstream_numbers[-1]

  latest_version_less_than_request = None
  latest_sha1_less_than_request = None
  for upstream_version, sha1 in upstream_versions:
    if upstream_version < version:
      latest_version_less_than_request = upstream_version
      latest_sha1_less_than_request = sha1

  if os.path.exists(workdir):
    shutil.rmtree(workdir)

  os.makedirs(workdir)
  subprocess.check_call(["git", "init"], cwd=workdir)
  subprocess.check_call(["git", "remote", "add", "origin", remote], cwd=workdir)
  subprocess.check_call(
      ["git", "fetch", "origin", "master"], cwd=workdir)

  if latest_version_less_than_request != latest_upstream_version:
    logger.warning(
        "Desired version tag %s is behind the latest deployed version %s."
        "I will try to find a suitable commit on master from which to branch.",
        version, latest_upstream_version)

    base_commit = subprocess.check_output(
        ["git", "merge-base", "master", latest_sha1_less_than_request]
    ).decode("utf-8").strip()
  else:
    base_commit = "master"

  subprocess.check_call(
      ["git", "checkout", base_commit], cwd=workdir)
  branch_name = "work-v{}".format(version)
  subprocess.check_call(
      ["git", "checkout", "-b", branch_name], cwd=workdir)

  filepath = os.path.join(workdir, "setup.py")
  content = SETUP_PY.format(versionstr)
  with io.open(filepath, "w", encoding="utf-8") as outfile:
    outfile.write(content)

  subprocess.check_call(["git", "add", "setup.py"], cwd=workdir)
  subprocess.check_call(
      ["git", "commit", "-m", "update to {}".format(versionstr)], cwd=workdir)
  subprocess.check_call(
      ["git", "tag", "v" + versionstr], cwd=workdir)

  if dry_run:
    logging.info("Skiping push of tag %s", "v" + versionstr)
  else:
    subprocess.check_call(
        ["git", "push", "origin", "v" + versionstr], cwd=workdir,
        env=get_env(deploykey_path))

  if keep:
    logging.info("Leaving work tree at %s", workdir)
  else:
    shutil.rmtree(workdir)
  return 0


def main():
  logging.basicConfig(level=logging.INFO)

  default_workdir = os.path.join(tempfile.gettempdir(), "scratch-work")
  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument("--version", help="Which version")
  argparser.add_argument(
      "--workdir", default=default_workdir,
      help="Directory we can use for work")
  argparser.add_argument(
      "--dry-run", action="store_true", help="Don't push when done")
  argparser.add_argument(
      "--keep", action="store_true", help="keep the worktree around")
  argparser.add_argument(
      "--deploykey", help="path to the deploykey", nargs="?")
  args = argparser.parse_args()

  version = args.version
  if version == "from-travis-tag":
    version = os.environ["TRAVIS_TAG"]
  return inner_main(
      args.workdir, version, args.deploykey, args.dry_run, args.keep)


if __name__ == "__main__":
  sys.exit(main())



# -*- coding: utf-8 -*-
# pylint: disable=too-many-lines
from __future__ import print_function
from __future__ import unicode_literals

import collections
import logging

from cmakelang.common import InternalError
from cmakelang import lex
from cmakelang.parse.util import (
    COMMENT_TOKENS, ONOFF_TOKENS, WHITESPACE_TOKENS
)
from cmakelang.parse.common import (
    FlowType, NodeType, TreeNode
)
from cmakelang.parse.simple_nodes import (
    CommentNode, OnOffNode, WhitespaceNode
)
from cmakelang.parse.statement_node import (
    AtWordStatementNode, StatementNode
)

logger = logging.getLogger(__name__)


class BodyNode(TreeNode):
  """Top level scope node contains all block-level nodes."""

  def __init__(self):
    super(BodyNode, self).__init__(NodeType.BODY)

  @classmethod
  def consume(cls, ctx, tokens, breakset=None):
    """
    Consume tokens and return a tree of nodes. Top-level consumer parsers
    comments, whitespace, statements, and flow control blocks.
    """
    if breakset is None:
      breakset = ()

    tree = cls()
    blocks = tree.children

    while tokens:
      token = tokens[0]
      if token.type in WHITESPACE_TOKENS:
        node = WhitespaceNode.consume(ctx, tokens)
        blocks.append(node)
      elif token.type in COMMENT_TOKENS:
        node = CommentNode.consume(ctx, tokens)
        blocks.append(node)
      elif token.type in ONOFF_TOKENS:
        node = OnOffNode.consume(ctx, tokens)
        blocks.append(node)
      elif token.type == lex.TokenType.BRACKET_COMMENT:
        node = CommentNode.consume(ctx, tokens)
        blocks.append(node)
      elif token.type == lex.TokenType.WORD:
        upper = token.spelling.upper()
        if upper in breakset:
          return tree
        if FlowType.get(upper) is not None:
          subtree = FlowControlNode.consume(ctx, tokens)
          blocks.append(subtree)
        else:
          subtree = StatementNode.consume(ctx, tokens)
          blocks.append(subtree)
      elif token.type == lex.TokenType.ATWORD:
        subtree = AtWordStatementNode.consume(ctx, tokens)
        blocks.append(subtree)
      elif token.type == lex.TokenType.BYTEORDER_MARK:
        tokens.pop(0)
      else:
        raise InternalError(
            "Unexpected {} token at {}:{}"
            .format(tokens[0].type.name,
                    tokens[0].begin.line, tokens[0].begin.col))

    return tree


FlowControlBlock = collections.namedtuple(
    "FlowControlBlock", ["open_stmt", "body", "close_stmt"])


class FlowControlNode(TreeNode):
  """Parent node of a flow-control statement set, including at least one opening
     statement, a body, and a closing statement. if() tree's may contain
     multiple opening statement and body pairs.
  """

  def __init__(self):
    super(FlowControlNode, self).__init__(NodeType.FLOW_CONTROL)
    self.blocks = []

  def get_block_with(self, node):
    """Return the block which contains the given node"""
    for block in self.blocks:
      if node in block:
        return block
    return None

  @classmethod
  def consume(cls, ctx, tokens):
    """
    Consume tokens and return a flow control tree. A flow control tree starts
    and ends with a statement node, and contains one or more body nodes. An if
    statement is the only flow control that includes multiple body nodes and
    they are separated by either else or elseif nodes.
    """
    flowtype = FlowType.from_name(tokens[0].spelling.upper())
    if flowtype is FlowType.IF:
      return IfBlockNode.consume(ctx, tokens)

    node = cls()
    endflow = 'END' + flowtype.name

    open_stmt = StatementNode.consume(ctx, tokens)
    node.children.append(open_stmt)
    body = BodyNode.consume(ctx, tokens, (endflow,))
    node.children.append(body)
    close_stmt = StatementNode.consume(ctx, tokens)
    node.children.append(close_stmt)

    node.blocks.append(FlowControlBlock(open_stmt, body, close_stmt))
    return node


class IfBlockNode(FlowControlNode):
  """IfBlocks are different than other `FlowControlNode` s in that they can
     contain multiple bodies.
  """

  @classmethod
  def consume(cls, ctx, tokens):
    """
    Consume tokens and return a flow control tree. ``IF`` statements are special
    because they have interior ``ELSIF`` and ``ELSE`` blocks, while all other
    flow control have a single body.
    """

    tree = cls()
    breakset = ("ELSE", "ELSEIF", "ENDIF")

    prev_open_stmt = None
    prev_body = None
    while tokens and tokens[0].spelling.upper() != "ENDIF":
      open_stmt = StatementNode.consume(ctx, tokens)
      body = BodyNode.consume(ctx, tokens, breakset)
      tree.children.append(open_stmt)
      tree.children.append(body)
      if prev_open_stmt is not None:
        tree.blocks.append(
            FlowControlBlock(prev_open_stmt, prev_body, open_stmt))

      prev_open_stmt = open_stmt
      prev_body = body

    if tokens:
      close_stmt = StatementNode.consume(ctx, tokens)
      tree.children.append(close_stmt)
      tree.blocks.append(
          FlowControlBlock(prev_open_stmt, prev_body, close_stmt))

    return tree



import json
import os
import re
import unittest


class TestVersionNumber(unittest.TestCase):
  """
  Verify that various documentation and configuration files are all using the
  same version number.
  """

  def __init__(self, *args):
    super(TestVersionNumber, self).__init__(*args)
    self.init_version = None
    self.repodir = None

  def setUp(self):
    thisdir = os.path.dirname(os.path.realpath(__file__))
    parent = os.path.dirname(thisdir)
    self.repodir = os.path.dirname(parent)
    with open(os.path.join(parent, "__init__.py")) as infile:
      initpy = infile.read()
    match = re.search("__version__ = ['\"]([^'\"]+)['\"]", initpy)
    if not match:
      self.fail("No version in __init__.py")
      return
    parts = match.group(1).split(".")
    self.init_version = ".".join(parts[:3])

  def test_install_documentation(self):
    filepath = os.path.join(self.repodir, "cmakelang/doc/installation.rst")
    with open(filepath) as infile:
      content = infile.read()

    match = re.search(r"pip install v(\S+).tar.gz", content)
    if not match:
      self.fail("Couldn't find 'pip install' in installation.rst")
      return

    self.assertEqual(self.init_version, match.group(1))

  def test_precommit_documentation(self):
    filepath = os.path.join(self.repodir, "cmakelang/doc/installation.rst")
    with open(filepath) as infile:
      content = infile.read()

    match = re.search(r"rev: v(\S+)", content)
    if not match:
      self.fail("Couldn't find 'rev:' in installation.rst")
      return

    self.assertEqual(self.init_version, match.group(1))

  def test_vscode_package_json(self):
    filepath = os.path.join(
        self.repodir, "cmakelang/vscode_extension/package.json")
    with open(filepath) as infile:
      data = json.load(infile)

    self.assertIn("version", data)
    json_version = data["version"]
    if "-" in json_version:
      json_version = json_version.split("-", 1)[0]
    self.assertEqual(json_version, self.init_version)

  def test_vscode_packagelock_json(self):
    filepath = os.path.join(
        self.repodir, "cmakelang/vscode_extension/package.json")
    with open(filepath) as infile:
      data = json.load(infile)

    self.assertIn("version", data)
    json_version = data["version"]
    if "-" in json_version:
      json_version = json_version.split("-", 1)[0]
    self.assertEqual(json_version, self.init_version)

  def test_changelog(self):
    """
    Ensure that the changelog includes an section for this version
    """
    version_str = "v" + self.init_version
    hruler = '-' * len(version_str)
    expect_str = version_str + "\n" + hruler

    filepath = os.path.join(
        self.repodir, "cmakelang/doc/changelog.rst")
    with open(filepath) as infile:
      content = infile.read()
    self.assertIn(expect_str, content)

  def test_relnotes(self):
    """
    Ensure that the release notes includes an section for this version
    """
    version_str = "v" + self.init_version
    hruler = '-' * len(version_str)
    expect_str = hruler + "\n" + version_str + "\n" + hruler

    filepath = os.path.join(
        self.repodir, "cmakelang/doc/release_notes.rst")
    with open(filepath) as infile:
      content = infile.read()
    self.assertIn(expect_str, content)


if __name__ == "__main__":
  unittest.main()



from __future__ import print_function, unicode_literals

import sys
import unittest

# pylint: disable=W0401,W0611,W0614
from cmakelang.lint.test import genfiles
from cmakelang.lint.test.expect_tests import gen_test_classes, ConfigTestCase
from cmakelang.lint.test.execution_tests import TestFormatFiles

if __name__ == "__main__":
  classnames = [
      "ConfigTestCase",
      "TestFormatFiles",
  ]

  classobj = None
  for classobj in gen_test_classes():
    globals()[classobj.__name__] = classobj
    classnames.append(classobj.__name__)
  del classobj

  if len(sys.argv) > 1 and sys.argv[1] == "print-names":
    print("\n".join(sorted(classnames)))
    sys.exit(0)

  unittest.main()



# -*- coding: utf-8 -*-
# pylint: disable=bad-continuation
# pylint: disable=too-many-lines
from __future__ import unicode_literals

import io
import os
import unittest

from cmakelang import configuration
from cmakelang.command_tests import assert_format, TestBase


class TestMiscFormatting(TestBase):
  """
  Ensure that various inputs format the way we want them to
  """
  kExpectNumSidecarTests = 86

  def test_config_hashruler_minlength(self):

    # make sure changing hashruler_min_length works correctly
    # self.config.enable_markup = False
    for min_width in {3, 5, 7, 9}:
      self.config.markup.hashruler_min_length = min_width

      # NOTE(josh): these tests use short rulers that wont be picked up by
      # the default pattern
      self.config.markup.ruler_pattern = (r'#{%d}#*' % (min_width - 1))

      just_shy = '#' * (min_width - 1)
      just_right = '#' * min_width
      longer = '#' * (min_width + 2)
      full_line = '#' * (self.config.format.line_width - 2)

      assert_format(self, """
# A comment: min_width={min_width}, just_shy
{just_shy}
""".format(min_width=min_width, just_shy=just_shy),
          """\
# A comment: min_width={min_width}, just_shy
#
""".format(min_width=min_width))

      assert_format(self, """
# A comment: min_width={min_width}, just_right
{just_right}
""".format(min_width=min_width, just_right=just_right),
          """\
# A comment: min_width={min_width}, just_right
# {full_line}
""".format(min_width=min_width, full_line=full_line))

      assert_format(self, """
# A comment: min_width={min_width}, longer
{longer}
""".format(min_width=min_width, longer=longer),
          """\
# A comment: min_width={min_width}, longer
# {full_line}
""".format(min_width=min_width, full_line=full_line))

  def test_windows_line_endings_input(self):
    self.source_str = (
        "#[[*********************************************\r\n"
        "* Information line 1\r\n"
        "* Information line 2\r\n"
        "************************************************]]\r\n")

    self.expect_format = """\
#[[*********************************************
* Information line 1
* Information line 2
************************************************]]
"""

  def test_windows_line_endings_output(self):
    config_dict = self.config.as_dict()
    config_dict['line_ending'] = 'windows'
    self.config = configuration.Configuration(**config_dict)

    self.source_str = """\
#[[*********************************************
* Information line 1
* Information line 2
************************************************]]"""

    self.expect_format = (
        "#[[*********************************************\r\n"
        "* Information line 1\r\n"
        "* Information line 2\r\n"
        "************************************************]]\r\n")

  def test_auto_line_endings(self):
    config_dict = self.config.as_dict()
    config_dict['line_ending'] = 'auto'
    self.config = configuration.Configuration(**config_dict)

    self.source_str = (
        "#[[*********************************************\r\n"
        "* Information line 1\r\n"
        "* Information line 2\r\n"
        "************************************************]]\r\n")

    self.expect_format = (
        "#[[*********************************************\r\n"
        "* Information line 1\r\n"
        "* Information line 2\r\n"
        "************************************************]]\r\n")

  def test_byte_order_mark(self):
    self.source_str = """\
\ufeffcmake_minimum_required(VERSION 2.8.11)
project(cmakelang_test)
"""
    self.expect_format = """\
cmake_minimum_required(VERSION 2.8.11)
project(cmakelang_test)
"""

    self.config.encode.emit_byteorder_mark = True
    self.source_str = """\
cmake_minimum_required(VERSION 2.8.11)
project(cmakelang_test)
"""
    self.expect_format = """\
\ufeffcmake_minimum_required(VERSION 2.8.11)
project(cmakelang_test)
"""

  def test_example_file(self):
    thisdir = os.path.dirname(__file__)
    infile_path = os.path.join(
        thisdir, '..', 'format/testdata', 'test_in.cmake')
    outfile_path = os.path.join(
        thisdir, '..', 'format/testdata', 'test_out.cmake')

    with io.open(infile_path, 'r', encoding='utf8') as infile:
      infile_text = infile.read()
    with io.open(outfile_path, 'r', encoding='utf8') as outfile:
      outfile_text = outfile.read()

    self.source_str = infile_text
    self.expect_format = outfile_text


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
# pylint: disable=W0613
from __future__ import print_function
from __future__ import unicode_literals

import logging

from cmakelang import lex
from cmakelang.common import InternalError
from cmakelang.parse.printer import dump_tree_tostr
from cmakelang.parse.util import (
    comment_belongs_up_tree,
    get_normalized_kwarg,
    get_tag,
    npargs_is_exact,
    pargs_are_full,
    IMPLICIT_PARG_TYPES,
    CommandSpec,
    PositionalSpec,
    WHITESPACE_TOKENS,
    should_break,
)
from cmakelang.parse.common import (
    NodeType, ParenBreaker, KwargBreaker, TreeNode
)
from cmakelang.parse.simple_nodes import CommentNode, OnOffNode

logger = logging.getLogger(__name__)


DEFAULT_PSPEC = PositionalSpec("*")


class ArgGroupNode(TreeNode):
  """A group-of-groups. Top-level for the argument subtree of a statement, as
     well as the argument subtree of a keyword."""

  def __init__(self):
    super(ArgGroupNode, self).__init__(NodeType.ARGGROUP)
    self.spec = None


class StandardArgTree(ArgGroupNode):
  """Argument tree for most cmake-statement commands. Generically arguments
     are composed of a positional argument list, followed by one or more
     keyword arguments, followed by one or more flags::

      command_name(parg1 parg2 parg3...
              KEYWORD1 kwarg1 kwarg2...
              KEYWORD2 kwarg3 kwarg4...
              FLAG1 FLAG2 FLAG3)

  """

  def __init__(self):
    super(StandardArgTree, self).__init__()
    self.parg_groups = []
    self.kwarg_groups = []
    self.cmdspec = None

  def check_required_kwargs(self, lint_ctx, required_kwargs):
    for kwargnode in self.kwarg_groups:
      required_kwargs.pop(get_normalized_kwarg(kwargnode.keyword.token), None)

    if required_kwargs:
      location = self.get_location()
      for token in self.get_semantic_tokens():
        location = token.get_location()
        break

      # NOTE(josh): inner sorted() needed for stable sort.
      missing_kwargs = sorted(
          (lintid, word) for word, lintid in sorted(required_kwargs.items()))
      for lintid, word in missing_kwargs:
        lint_ctx.record_lint(lintid, word, location=location)

  @classmethod
  def parse2(cls, ctx, tokens, cmdspec, kwargs, breakstack):
    """
    Standard parser for the commands in the form of::

        command_name(parg1 parg2 parg3...
                    KEYWORD1 kwarg1 kwarg2...
                    KEYWORD2 kwarg3 kwarg4...
                    FLAG1 FLAG2 FLAG3)
    The parser starts off as a positional parser. If a keyword or flag is
    encountered the positional parser is popped off the parse stack. If it was
    a keyword then the keyword parser is pushed on the parse stack. If it was
    a flag than a new flag parser is pushed onto the stack.
    """

    # NOTE(josh): we will pop things off this list, so let's make a copy
    pargspecs = list(cmdspec.pargs)
    tree = cls()
    tree.cmdspec = cmdspec

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    while tokens and tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # NOTE(josh): if there is only one non-exact legacy specification then we
    # reuse that specification for any additional positional arguments that we
    # pick up. This is to maintain the current/legacy behavior of simple
    # positional argument specifications
    # TODO(josh): double check the reasoning for this. I think it might be
    # mistaken and unnecessary
    default_spec = DEFAULT_PSPEC
    if (len(pargspecs) == 1 and pargspecs[0].legacy
        and not npargs_is_exact(pargspecs[0].nargs)):
      default_spec = pargspecs.pop(0)

    all_flags = list(default_spec.flags)
    for pspec in pargspecs:
      all_flags.extend(pspec.flags)

    kwarg_breakstack = breakstack + [
        KwargBreaker(list(kwargs.keys()) + all_flags)]

    while tokens:
      # If it is a whitespace token then put it directly in the parse tree at
      # the current depth
      if tokens[0].type in WHITESPACE_TOKENS:
        tree.children.append(tokens.pop(0))
        continue

      # If it's a comment, then add it at the current depth
      if tokens[0].type in (lex.TokenType.COMMENT,
                            lex.TokenType.BRACKET_COMMENT):
        if comment_belongs_up_tree(ctx, tokens, tree, breakstack):
          break
        tree.children.append(CommentNode.consume(ctx, tokens))
        continue

      # If it's a sentinel comment, then add it at the current depth
      if tokens[0].type in (lex.TokenType.FORMAT_OFF,
                            lex.TokenType.FORMAT_ON):
        tree.children.append(OnOffNode.consume(ctx, tokens))
        continue

      # Break if the next token belongs to a parent parser, i.e. if it
      # matches a keyword argument of something higher in the stack, or if
      # it closes a parent group.
      if should_break(tokens[0], breakstack):
        # NOTE(josh): if spec.nargs is an exact number of arguments, then we
        # shouldn't break on kwarg match from a parent parser. Instead, we
        # should consume that many tokens. This is a hack to deal with
        # ```install(RUNTIME COMPONENT runtime)``. In this case the second
        # occurance of "runtime" should not match the ``RUNTIME`` keyword
        # and should not break the positional parser.
        # TODO(josh): this is kind of hacky because it will force the positional
        # parser to consume a right parenthesis and will lead to parse errors
        # in the event of a missing positional argument. Such errors will be
        # difficult to debug for the user.
        if pargspecs:
          pspec = pargspecs[0]
        else:
          pspec = default_spec

        if not npargs_is_exact(pspec.nargs) or pspec.nargs == 0:
          break

      ntokens = len(tokens)
      word = get_normalized_kwarg(tokens[0])
      if word in kwargs:
        with ctx.pusharg(tree):
          subtree = KeywordGroupNode.parse(
              ctx, tokens, word, kwargs[word], kwarg_breakstack)
        tree.kwarg_groups.append(subtree)
      else:
        if pargspecs:
          pspec = pargspecs.pop(0)
        else:
          pspec = default_spec

        other_flags = []
        for otherspec in pargspecs:
          for flag in otherspec.flags:
            if flag in pspec.flags:
              continue
            other_flags.append(flag)
        positional_breakstack = breakstack + [
            KwargBreaker(list(kwargs.keys()) + other_flags)]

        with ctx.pusharg(tree):
          subtree = PositionalGroupNode.parse2(
              ctx, tokens, pspec, positional_breakstack)
          tree.parg_groups.append(subtree)

      if len(tokens) >= ntokens:
        raise InternalError(
            "parsed an empty subtree at {}:\n  {}\n pspec: {}"
            .format(tokens[0], dump_tree_tostr([tree]), pspec))
      tree.children.append(subtree)
    return tree

  @classmethod
  def parse(cls, ctx, tokens, npargs, kwargs, flags, breakstack):
    """
    Standard parser for the commands in the form of::

        command_name(parg1 parg2 parg3...
                    KEYWORD1 kwarg1 kwarg2...
                    KEYWORD2 kwarg3 kwarg4...
                    FLAG1 FLAG2 FLAG3)

    The parser starts off as a positional parser. If a keyword or flag is
    encountered the positional parser is popped off the parse stack. If it was
    a keyword then the keyword parser is pushed on the parse stack. If it was
    a flag than a new flag parser is pushed onto the stack.
    """
    if isinstance(npargs, IMPLICIT_PARG_TYPES):
      pargspecs = [PositionalSpec(npargs, flags=flags, legacy=True)]
    else:
      assert isinstance(npargs, (list, tuple)), (
          "Invalid positional group specification of type {}"
          .format(type(npargs).__name__))
      assert flags is None, (
          "Invalid usage of old-style 'flags' parameter with new style"
          " positional group specifications")
      pargspecs = npargs

    return cls.parse2(
        ctx, tokens, CommandSpec("<none>", pargspecs), kwargs, breakstack)


class StandardParser(object):
  def __init__(self, npargs=None, kwargs=None, flags=None, doc=None):
    if npargs is None:
      npargs = "*"
    if flags is None:
      flags = []
    if kwargs is None:
      kwargs = {}

    self.npargs = npargs
    self.kwargs = kwargs
    self.flags = flags
    self.doc = doc

  def __call__(self, ctx, tokens, breakstack):
    return StandardArgTree.parse(
        ctx, tokens, self.npargs, self.kwargs, self.flags, breakstack)


class StandardParser2(object):
  def __init__(self, cmdspec=None, funtree=None, doc=None):
    if cmdspec is None:
      cmdspec = ("<none>", "*")
    if funtree is None:
      funtree = {}

    self.cmdspec = cmdspec
    self.funtree = funtree
    self.doc = doc

  @property
  def pspec(self):
    return self.cmdspec.pargs

  @property
  def kwargs(self):
    return self.funtree

  def __call__(self, ctx, tokens, breakstack):
    return StandardArgTree.parse2(
        ctx, tokens, self.cmdspec, self.funtree, breakstack)


class KeywordNode(TreeNode):
  """Node that stores a single keyword token and possibly it's associated
     comments."""

  def __init__(self):
    super(KeywordNode, self).__init__(NodeType.KEYWORD)
    self.token = None

  @classmethod
  def parse(cls, _ctx, tokens):
    node = cls()
    node.token = tokens.pop(0)
    node.children.append(node.token)
    # TODO(josh)[c490dba]: allow keywords to have trailing comments
    # consume_trailing_comment(node, tokens)
    return node


class KeywordGroupNode(TreeNode):
  """Argument subtree for a keyword and its arguments."""

  def __init__(self):
    super(KeywordGroupNode, self).__init__(NodeType.KWARGGROUP)
    self.keyword = None
    self.body = None

  @classmethod
  def parse(cls, ctx, tokens, word, subparser, breakstack):
    """
    Parse a standard `KWARG arg1 arg2 arg3...` style keyword argument list.
    """
    assert tokens[0].spelling.upper() == word.upper(), \
        "somehow dispatched wrong kwarg parse"

    tree = cls()
    keyword = KeywordNode.parse(ctx, tokens)
    tree.keyword = keyword
    tree.children.append(keyword)

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    while tokens and tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))

    ntokens = len(tokens)
    with ctx.pusharg(tree):
      subtree = subparser(ctx, tokens, breakstack)
    if len(tokens) < ntokens:
      tree.body = subtree
      tree.children.append(subtree)
    return tree


class PositionalGroupNode(TreeNode):
  """Argument subtree for one or more single positional argument tokens."""

  def __init__(self, sortable=False, tags=None):
    super(PositionalGroupNode, self).__init__(NodeType.PARGGROUP)
    self.sortable = sortable
    self.spec = None
    self.tags = []
    if tags is not None:
      self.tags.extend(tags)

  @classmethod
  def parse2(cls, ctx, tokens, spec, breakstack):
    """
    Parse a continuous sequence of `npargs` positional arguments. If npargs is
    an integer we will consume exactly that many arguments. If it is not an
    integer then it is a string meaning:

    * "?": zero or one
    * "*": zero or more
    * "+": one or more
    """

    tree = cls(sortable=spec.sortable, tags=spec.tags)
    tree.spec = spec
    nconsumed = 0

    # Strip off any preceeding whitespace (note that in most cases this has
    # already been done but in some cases (such ask kwarg subparser) where
    # it hasn't
    while tokens and tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))

    # If the first non-whitespace token is a cmake-format tag annotating
    # sortability, then parse it out here and record the annotation
    if tokens and get_tag(tokens[0]) in ("sortable", "sort"):
      tree.sortable = True
    elif tokens and get_tag(tokens[0]) in ("unsortable", "unsort"):
      tree.sortable = False

    while tokens:

      # Break if we have consumed   enough positional arguments
      if pargs_are_full(spec.nargs, nconsumed):
        break

      # Break if the next token belongs to a parent parser, i.e. if it
      # matches a keyword argument of something higher in the stack, or if
      # it closes a parent group.
      if should_break(tokens[0], breakstack):
        # NOTE(josh): if spec.nargs is an exact number of arguments, then we
        # shouldn't break on kwarg match from a parent parser. Instead, we
        # should consume the token. This is a hack to deal with
        # ```install(RUNTIME COMPONENT runtime)``. In this case the second
        # occurance of "runtime" should not match the ``RUNTIME`` keyword
        # and should not break the positional parser.
        # TODO(josh): this is kind of hacky because it will force the positional
        # parser to consume a right parenthesis and will lead to parse errors
        # in the event of a missing positional argument. Such errors will be
        # difficult to debug for the user.
        if not npargs_is_exact(spec.nargs):
          break

        if tokens[0].type == lex.TokenType.RIGHT_PAREN:
          break

      # If this is the start of a parenthetical group, then parse the group
      # NOTE(josh): syntatically this probably shouldn't be allowed here, but
      # cmake seems to accept it so we probably should too.
      if tokens[0].type == lex.TokenType.LEFT_PAREN:
        with ctx.pusharg(tree):
          subtree = ParenGroupNode.parse(ctx, tokens, breakstack)
        tree.children.append(subtree)
        continue

      # If it is a whitespace token then put it directly in the parse tree at
      # the current depth
      if tokens[0].type in WHITESPACE_TOKENS:
        tree.children.append(tokens.pop(0))
        continue

      # If it's a comment token not associated with an argument, then put it
      # directly into the parse tree at the current depth
      if tokens[0].type in (lex.TokenType.COMMENT,
                            lex.TokenType.BRACKET_COMMENT):
        if comment_belongs_up_tree(ctx, tokens, tree, breakstack):
          break

        child = CommentNode.consume(ctx, tokens)
        tree.children.append(child)
        continue

      # If it's a sentinel comment, then add it at the current depth
      if tokens[0].type in (lex.TokenType.FORMAT_OFF,
                            lex.TokenType.FORMAT_ON):
        tree.children.append(OnOffNode.consume(ctx, tokens))
        continue

      # Otherwise is it is a positional argument, so add it to the tree as such
      if get_normalized_kwarg(tokens[0]) in spec.flags:
        child = TreeNode(NodeType.FLAG)
      else:
        child = TreeNode(NodeType.ARGUMENT)

      child.children.append(tokens.pop(0))
      CommentNode.consume_trailing(ctx, tokens, child)
      tree.children.append(child)
      nconsumed += 1

    return tree

  @classmethod
  def parse(cls, ctx, tokens, npargs, flags, breakstack, sortable=False,
            tags=None):
    """
    Parse a continuous sequence of `npargs` positional arguments. If npargs is
    an integer we will consume exactly that many arguments. If it is not an
    integer then it is a string meaning:

    * "?": zero or one
    * "*": zero or more
    * "+": one or more
    """

    spec = PositionalSpec(npargs, sortable, tags, flags)
    return cls.parse2(ctx, tokens, spec, breakstack)


class PositionalParser(object):
  def __init__(self, npargs=None, flags=None, sortable=False):
    if npargs is None:
      npargs = "*"
    if flags is None:
      flags = []

    self.npargs = npargs
    self.flags = flags
    self.sortable = sortable

  def __call__(self, ctx, tokens, breakstack):
    return PositionalGroupNode.parse(
        ctx, tokens, self.npargs, self.flags, breakstack, self.sortable)


class ParenGroupNode(TreeNode):
  """Argument subtree for a parenthetical group."""

  def __init__(self):
    super(ParenGroupNode, self).__init__(NodeType.PARENGROUP)

  @classmethod
  def parse(cls, ctx, tokens, _breakstack):
    """
    Consume a parenthetical group of arguments from `tokens` and return the
    parse subtree rooted at this group.  `argstack` contains a stack of all
    early break conditions that are currently "opened".
    """

    assert tokens[0].type == lex.TokenType.LEFT_PAREN
    tree = cls()
    lparen = TreeNode(NodeType.LPAREN)
    lparen.children.append(tokens.pop(0))
    tree.children.append(lparen)

    with ctx.pusharg(tree):
      subtree = ConditionalGroupNode.parse(ctx, tokens, [ParenBreaker()])
    tree.children.append(subtree)

    if tokens[0].type != lex.TokenType.RIGHT_PAREN:
      raise ValueError(
          "Unexpected {} token at {}, expecting r-paren, got {}"
          .format(tokens[0].type.name, tokens[0].get_location(),
                  tokens[0].content))
    rparen = TreeNode(NodeType.RPAREN)
    rparen.children.append(tokens.pop(0))
    tree.children.append(rparen)

    # NOTE(josh): parenthetical groups can have trailing comments because
    # they have closing punctuation
    CommentNode.consume_trailing(ctx, tokens, tree)

    return tree


CONDITIONAL_FLAGS = [
    "COMMAND",
    "DEFINED",
    "EQUAL",
    "EXISTS",
    "GREATER",
    "LESS",
    "IS_ABSOLUTE",
    "IS_DIRECTORY",
    "IS_NEWER_THAN",
    "IS_SYMLINK",
    "MATCHES",
    "NOT",
    "POLICY",
    "STRLESS",
    "STRGREATER",
    "STREQUAL",
    "TARGET",
    "TEST",
    "VERSION_EQUAL",
    "VERSION_GREATER",
    "VERSION_LESS",
]


class ConditionalGroupNode(ArgGroupNode):
  @classmethod
  def parse(cls, ctx, tokens, breakstack):
    """
    Parser for the commands that take conditional arguments. Similar to the
    standard parser but it understands parentheses and can generate
    parenthentical groups::

        while(CONDITION1 AND (CONDITION2 OR CONDITION3)
              OR (CONDITION3 AND (CONDITION4 AND CONDITION5)
              OR CONDITION6)
    """
    kwargs = {
        'AND': cls.parse,
        'OR': cls.parse
    }
    flags = list(CONDITIONAL_FLAGS)
    tree = cls()

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    while tokens and tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    flags = [flag.upper() for flag in flags]
    breaker = KwargBreaker(list(kwargs.keys()))
    child_breakstack = breakstack + [breaker]

    while tokens:
      # Break if the next token belongs to a parent parser, i.e. if it
      # matches a keyword argument of something higher in the stack, or if
      # it closes a parent group.
      if should_break(tokens[0], breakstack):
        break

      # If it is a whitespace token then put it directly in the parse tree at
      # the current depth
      if tokens[0].type in WHITESPACE_TOKENS:
        tree.children.append(tokens.pop(0))
        continue

      # If it's a comment, then add it at the current depth
      if tokens[0].type in (lex.TokenType.COMMENT,
                            lex.TokenType.BRACKET_COMMENT):
        # TODO(josh): not sure if we should check comment_belongs_up_tree
        # here
        child = CommentNode()
        tree.children.append(child)
        child.children.append(tokens.pop(0))
        continue

      # If it's a sentinel comment, then add it at the current depth
      if tokens[0].type in (lex.TokenType.FORMAT_OFF,
                            lex.TokenType.FORMAT_ON):
        tree.children.append(OnOffNode.consume(ctx, tokens))
        continue

      # If this is the start of a parenthetical group, then parse the group
      if tokens[0].type == lex.TokenType.LEFT_PAREN:
        with ctx.pusharg(tree):
          subtree = ParenGroupNode.parse(ctx, tokens, breakstack)
        tree.children.append(subtree)
        continue

      ntokens = len(tokens)
      word = get_normalized_kwarg(tokens[0])
      if word in kwargs:
        with ctx.pusharg(tree):
          subtree = KeywordGroupNode.parse(
              ctx, tokens, word, kwargs[word], child_breakstack)
        assert len(tokens) < ntokens, "parsed an empty subtree"
        tree.children.append(subtree)
        continue

      # Otherwise is it is a positional argument, so add it to the tree as such
      with ctx.pusharg(tree):
        child = PositionalGroupNode.parse(
            ctx, tokens, '+', flags, child_breakstack)
      # token = tokens.pop(0)
      # if get_normalized_kwarg(token) in flags:
      #   child = TreeNode(NodeType.FLAG)
      # else:
      #   child = TreeNode(NodeType.ARGUMENT)

      # child.children.append(token)
      # consume_trailing_comment(child, tokens)
      tree.children.append(child)

    return tree



from __future__ import unicode_literals


class EnumObject(object):
  """
  Simple enumeration base. Design inspired by clang python bindings
  BaseEnumeration. Subclasses must provide class member _id_map.
  """
  _id_map = {}

  @classmethod
  def register_value(cls, value, obj):
    assert cls._id_map is not EnumObject._id_map

    # If there is already an object with this value then raise an exception
    if value in cls._id_map:
      raise ValueError('{0} value {1} already loaded'.format(
          cls.__name__, value))

    cls._id_map[value] = obj

  @classmethod
  def from_id(cls, qid):
    if qid in cls._id_map:
      return cls._id_map[qid]

    raise ValueError("{} is not a valid {} value".format(qid, cls.__name__))

  @classmethod
  def assign_names(cls):
    for key, value in vars(cls).items():
      if isinstance(value, cls):
        value.name_ = key

  @classmethod
  def from_name(cls, name):
    if hasattr(cls, name):
      obj = getattr(cls, name)
      if isinstance(obj, cls):
        return obj
    raise ValueError("{} is not a valid {} enum".format(name, cls.__name__))

  @classmethod
  def get(cls, name, default=None):
    if hasattr(cls, name):
      obj = getattr(cls, name)
      if isinstance(obj, cls):
        return obj

      raise ValueError("{} is not a valid {} enum".format(name, cls.__name__))
    return default

  def __init__(self, value):
    self.value = value
    self.name_ = None
    self.__class__.register_value(value, self)

  @property
  def name(self):
    """Get the enumeration name of this value."""
    if self.name_ is None:
      self.__class__.assign_names()
    return self.name_

  def as_dict(self):
    return self.name

  def __repr__(self):
    return '{}.{}'.format(self.__class__.__name__, self.name)


def stable_wrap(wrapper, paragraph_text):
  """
  textwrap doesn't appear to be stable. We run it multiple times until it
  converges
  """

  history = [paragraph_text]
  prev_text = paragraph_text
  wrap_text = paragraph_text
  for _ in range(8):
    lines = wrapper.wrap(wrap_text)
    next_text = '\n'.join(line for line in lines)
    if next_text == prev_text:
      return lines
    prev_text = next_text
    history.append(next_text)

    lines[0] = lines[0][len(wrapper.initial_indent):]
    lines[1:] = [line[len(wrapper.subsequent_indent):] for line in lines[1:]]
    wrap_text = '\n'.join(lines)

  assert False, ("textwrap failed to converge on:\n\n {}"
                 .format('\n\n'.join(history)))
  return []


class UserError(Exception):
  """Raised when we encounter a fatal problem with usage: e.g. parse error,
     config error, input error, etc."""

  def __init__(self, msg=None):
    super(UserError, self).__init__()
    self.msg = msg

  def __repr__(self):
    return self.msg


class InternalError(Exception):
  """Raised when we encounter something we do not expect, indicating a problem
     with the code itself."""

  def __init__(self, msg=None):
    super(InternalError, self).__init__()
    self.msg = msg

  def __repr__(self):
    return self.msg


class FormatError(Exception):
  """Raised during format or format --check indicating that the program should
     exit with nonzero status code."""

  def __init__(self, msg=None):
    super(FormatError, self).__init__()
    self.msg = msg

  def __repr__(self):
    return self.msg



from cmakelang.parse.argument_nodes import (
    PositionalParser,
    StandardArgTree,
)
from cmakelang.parse.util import (
    iter_semantic_tokens,
)


def parse_ctest_build(ctx, tokens, breakstack):
  """
  ::

    ctest_build([BUILD <build-dir>] [APPEND]
                [CONFIGURATION <config>]
                [FLAGS <flags>]
                [PROJECT_NAME <project-name>]
                [TARGET <target-name>]
                [NUMBER_ERRORS <num-err-var>]
                [NUMBER_WARNINGS <num-warn-var>]
                [RETURN_VALUE <result-var>]
                [CAPTURE_CMAKE_ERROR <result-var>]
                )

  :see: https://cmake.org/cmake/help/latest/command/ctest_build.html
  """
  kwargs = {
      "BUILD": PositionalParser(1),
      "CONFIGURATION": PositionalParser(1),
      "FLAGS": PositionalParser(1),
      "PROJECT_NAME": PositionalParser(1),
      "TARGET": PositionalParser(1),
      "NUMBER_ERRORS": PositionalParser(1),
      "NUMBER_WARNINGS": PositionalParser(1),
      "RETURN_VALUE": PositionalParser(1),
      "CAPTURE_CMAKE_ERROR": PositionalParser(1)
  }
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, ["APPEND"], breakstack)


def parse_ctest_configure(ctx, tokens, breakstack):
  """
  ::

    ctest_configure([BUILD <build-dir>] [SOURCE <source-dir>] [APPEND]
                    [OPTIONS <options>] [RETURN_VALUE <result-var>] [QUIET]
                    [CAPTURE_CMAKE_ERROR <result-var>])

  :see: https://cmake.org/cmake/help/latest/command/ctest_configure.html
  """
  kwargs = {
      "BUILD": PositionalParser(1),
      "SOURCE": PositionalParser(1),
      "OPTIONS": PositionalParser(1),
      "RETURN_VALUE": PositionalParser(1),
      "CAPTURE_CMAKE_ERROR": PositionalParser(1)
  }
  return StandardArgTree.parse(
      ctx, tokens, "+", kwargs, ["APPEND", "QUIET"], breakstack)


def parse_ctest_coverage(ctx, tokens, breakstack):
  """
  ::

    ctest_coverage([BUILD <build-dir>] [APPEND]
                   [LABELS <label>...]
                   [RETURN_VALUE <result-var>]
                   [CAPTURE_CMAKE_ERROR <result-var>]
                   [QUIET]
                   )

  :see: https://cmake.org/cmake/help/latest/command/ctest_coverage.html
  """
  kwargs = {
      "BUILD": PositionalParser(1),
      "LABELS": PositionalParser("+"),
      "RETURN_VALUE": PositionalParser(1),
      "CAPTURE_CMAKE_ERROR": PositionalParser(1),
  }
  return StandardArgTree.parse(
      ctx, tokens, "+", kwargs, ["APPEND", "QUIET"], breakstack)


def parse_ctest_empty_binary_directory(ctx, tokens, breakstack):
  """
  ::

    ctest_empty_binary_directory( directory )

  :see: https://cmake.org/cmake/help/latest/command/ctest_empty_binary_directory.html
  """
  return StandardArgTree.parse(ctx, tokens, 1, {}, [], breakstack)


def parse_ctest_memcheck(ctx, tokens, breakstack):
  """
  ::

    ctest_memcheck([BUILD <build-dir>] [APPEND]
                   [START <start-number>]
                   [END <end-number>]
                   [STRIDE <stride-number>]
                   [EXCLUDE <exclude-regex>]
                   [INCLUDE <include-regex>]
                   [EXCLUDE_LABEL <label-exclude-regex>]
                   [INCLUDE_LABEL <label-include-regex>]
                   [EXCLUDE_FIXTURE <regex>]
                   [EXCLUDE_FIXTURE_SETUP <regex>]
                   [EXCLUDE_FIXTURE_CLEANUP <regex>]
                   [PARALLEL_LEVEL <level>]
                   [TEST_LOAD <threshold>]
                   [SCHEDULE_RANDOM <ON|OFF>]
                   [STOP_TIME <time-of-day>]
                   [RETURN_VALUE <result-var>]
                   [DEFECT_COUNT <defect-count-var>]
                   [QUIET]
                   )

  :see: https://cmake.org/cmake/help/latest/command/ctest_memcheck.html
  """
  kwargs = {
      "BUILD": PositionalParser(1),
      "START": PositionalParser(1),
      "END": PositionalParser(1),
      "STRIDE": PositionalParser(1),
      "EXCLUDE": PositionalParser(1),
      "INCLUDE": PositionalParser(1),
      "EXCLUDE_LABEL": PositionalParser(1),
      "INCLUDE_LABEL": PositionalParser(1),
      "EXCLUDE_FIXTURE": PositionalParser(1),
      "EXCLUDE_FIXTURE_SETUP": PositionalParser(1),
      "EXCLUDE_FIXTURE_CLEANUP": PositionalParser(1),
      "PARALLEL_LEVEL": PositionalParser(1),
      "TEST_LOAD": PositionalParser(1),
      "SCHEDULE_RANDOM": PositionalParser(1, flags=["ON", "OFF"]),
      "STOP_TIME": PositionalParser(1),
      "RETURN_VALUE": PositionalParser(1),
      "DEFECT_COUNT": PositionalParser(1),
  }
  flags = ["APPEND", "QUIET"]
  return StandardArgTree.parse(ctx, tokens, 0, kwargs, flags, breakstack)


def parse_ctest_read_custom_files(ctx, tokens, breakstack):
  """
  ::

    ctest_read_custom_files( directory ... )

  :see: https://cmake.org/cmake/help/latest/command/ctest_read_custom_files.html
  """
  return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)


def parse_ctest_run_script(ctx, tokens, breakstack):
  """
  ::

    ctest_run_script([NEW_PROCESS] script_file_name script_file_name1
                     script_file_name2 ... [RETURN_VALUE var])

  :see: https://cmake.org/cmake/help/latest/command/ctest_run_script.html
  """
  kwargs = {"RETURN_VALUE": PositionalParser(1)}
  flags = ["NEW_PROCESS"]
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, flags, breakstack)


def parse_ctest_sleep(ctx, tokens, breakstack):
  """
  ::

    ctest_sleep(<seconds>)
    ctest_sleep(<time1> <duration> <time2>)

  :see: https://cmake.org/cmake/help/latest/command/ctest_sleep.html
  """
  semantic_tokens = list(iter_semantic_tokens(tokens))
  if len(semantic_tokens) == 1:
    return StandardArgTree.parse(ctx, tokens, 1, {}, [], breakstack)
  return StandardArgTree.parse(ctx, tokens, 3, {}, [], breakstack)


def parse_ctest_start(ctx, tokens, breakstack):
  """
  ::

    ctest_start(<model> [<source> [<binary>]] [GROUP <group>] [QUIET])
    ctest_start([<model> [<source> [<binary>]]] [GROUP <group>] APPEND [QUIET])

  :see: https://cmake.org/cmake/help/latest/command/ctest_start.html
  """
  kwargs = {
      "GROUP": PositionalParser(1)
  }
  flags = ["QUIET", "APPEND"]
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, flags, breakstack)


def parse_ctest_submit(ctx, tokens, breakstack):
  """
  ::

    ctest_submit([PARTS <part>...] [FILES <file>...]
                 [SUBMIT_URL <url>]
                 [BUILD_ID <result-var>]
                 [HTTPHEADER <header>]
                 [RETRY_COUNT <count>]
                 [RETRY_DELAY <delay>]
                 [RETURN_VALUE <result-var>]
                 [CAPTURE_CMAKE_ERROR <result-var>]
                 [QUIET]
                 )

  :see: https://cmake.org/cmake/help/latest/command/ctest_submit.html
  """
  kwargs = {
      "PARTS": PositionalParser("+"),
      "FILES": PositionalParser("+"),
      "SUBMIT_URL": PositionalParser(1),
      "BUILD_ID": PositionalParser(1),
      "HTTPHEADER": PositionalParser(1),
      "RETRY_COUNT": PositionalParser(1),
      "RETRY_DELAY": PositionalParser(1),
      "RETURN_VALUE": PositionalParser(1),
      "CAPTURE_CMAKE_ERROR": PositionalParser(1),
  }
  flags = ["QUIET"]
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, flags, breakstack)


def parse_ctest_test(ctx, tokens, breakstack):
  """
  ::

    ctest_test([BUILD <build-dir>] [APPEND]
               [START <start-number>]
               [END <end-number>]
               [STRIDE <stride-number>]
               [EXCLUDE <exclude-regex>]
               [INCLUDE <include-regex>]
               [EXCLUDE_LABEL <label-exclude-regex>]
               [INCLUDE_LABEL <label-include-regex>]
               [EXCLUDE_FIXTURE <regex>]
               [EXCLUDE_FIXTURE_SETUP <regex>]
               [EXCLUDE_FIXTURE_CLEANUP <regex>]
               [PARALLEL_LEVEL <level>]
               [RESOURCE_SPEC_FILE <file>]
               [TEST_LOAD <threshold>]
               [SCHEDULE_RANDOM <ON|OFF>]
               [STOP_TIME <time-of-day>]
               [RETURN_VALUE <result-var>]
               [CAPTURE_CMAKE_ERROR <result-var>]
               [QUIET]
               )

  :see: https://cmake.org/cmake/help/latest/command/ctest_test.html
  """
  kwargs = {
      "BUILD": PositionalParser(1),
      "START": PositionalParser(1),
      "END": PositionalParser(1),
      "STRIDE": PositionalParser(1),
      "EXCLUDE": PositionalParser(1),
      "INCLUDE": PositionalParser(1),
      "EXCLUDE_LABEL": PositionalParser(1),
      "INCLUDE_LABEL": PositionalParser(1),
      "EXCLUDE_FIXTURE": PositionalParser(1),
      "EXCLUDE_FIXTURE_SETUP": PositionalParser(1),
      "EXCLUDE_FIXTURE_CLEANUP": PositionalParser(1),
      "PARALLEL_LEVEL": PositionalParser(1),
      "RESOURCE_SPEC_FILE": PositionalParser(1),
      "TEST_LOAD": PositionalParser(1),
      "SCHEDULE_RANDOM": PositionalParser(1, flags=["ON", "OFF"]),
      "STOP_TIME": PositionalParser(1),
      "RETURN_VALUE": PositionalParser(1),
      "CAPTURE_CMAKE_ERROR": PositionalParser(1),
  }
  flags = ["APPEND", "QUIET"]
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, flags, breakstack)


def parse_ctest_update(ctx, tokens, breakstack):
  """
  ::

    ctest_update([SOURCE <source-dir>]
                 [RETURN_VALUE <result-var>]
                 [CAPTURE_CMAKE_ERROR <result-var>]
                 [QUIET])

  :see: https://cmake.org/cmake/help/latest/command/ctest_update.html
  """
  kwargs = {
      "SOURCE": PositionalParser(1),
      "RETURN_VALUE": PositionalParser(1),
      "CAPTURE_CMAKE_ERROR": PositionalParser(1)
  }
  flags = ["QUIET"]
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, flags, breakstack)


def parse_ctest_upload(ctx, tokens, breakstack):
  """
  ::

    ctest_upload(FILES <file>... [QUIET] [CAPTURE_CMAKE_ERROR <result-var>])

  :see: https://cmake.org/cmake/help/latest/command/ctest_upload.html
  """
  kwargs = {
      "FILES": PositionalParser("+"),
      "CAPTURE_CMAKE_ERROR": PositionalParser(1)
  }
  flags = ["QUIET"]
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, flags, breakstack)


def parse_target_compile_definitions(ctx, tokens, breakstack):
  """
  ::

    target_compile_definitions(<target>
        <INTERFACE|PUBLIC|PRIVATE> [items1...]
        [<INTERFACE|PUBLIC|PRIVATE> [items2...] ...])

  :see: https://cmake.org/cmake/help/latest/command/target_compile_definitions.html
  """
  kwargs = {
      "INTERFACE": PositionalParser("+"),
      "PUBLIC": PositionalParser("+"),
      "PRIVATE": PositionalParser("+")
  }
  return StandardArgTree.parse(ctx, tokens, 1, kwargs, [], breakstack)


def parse_target_compile_features(ctx, tokens, breakstack):
  """
  ::

    target_compile_features(<target> <PRIVATE|PUBLIC|INTERFACE> <feature> [...])

  :see: https://cmake.org/cmake/help/latest/command/target_compile_features.html
  """
  kwargs = {
      "INTERFACE": PositionalParser("+"),
      "PUBLIC": PositionalParser("+"),
      "PRIVATE": PositionalParser("+"),
  }
  return StandardArgTree.parse(ctx, tokens, 1, kwargs, [], breakstack)


def parse_target_sources(ctx, tokens, breakstack):
  """
  ::

    target_sources(<target>
      <INTERFACE|PUBLIC|PRIVATE> [items1...]
      [<INTERFACE|PUBLIC|PRIVATE> [items2...] ...])

  :see: https://cmake.org/cmake/help/latest/command/target_sources.html
  """
  kwargs = {
      "INTERFACE": PositionalParser("+"),
      "PUBLIC": PositionalParser("+"),
      "PRIVATE": PositionalParser("+"),
  }
  return StandardArgTree.parse(ctx, tokens, 1, kwargs, [], breakstack)


def parse_unset(ctx, tokens, breakstack):
  """
  ::

    unset(<variable> [CACHE | PARENT_SCOPE])
    unset(ENV{<variable>})

  :see: https://cmake.org/cmake/help/latest/command/unset.html
  """
  flags = ["CACHE", "PARENT_SCOPE"]
  return StandardArgTree.parse(ctx, tokens, "+", {}, flags, breakstack)


def parse_variable_watch(ctx, tokens, breakstack):
  """
  ::

    variable_watch(<variable> [<command>])

  :see: https://cmake.org/cmake/help/latest/command/variable_watch.html
  """
  return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)


def parse_add_compile_definitions(ctx, tokens, breakstack):
  """
  ::

  add_compile_definitions(<definition> ...)

  :see: https://cmake.org/cmake/help/latest/command/add_compile_definitions.html
  """
  return StandardArgTree.parse(ctx, tokens, "+", {}, [], breakstack)


def populate_db(parse_db):
  parse_db["ctest_build"] = parse_ctest_build
  parse_db["ctest_configure"] = parse_ctest_configure
  parse_db["ctest_coverage"] = parse_ctest_coverage
  parse_db["ctest_empty_binary_directory"] = parse_ctest_empty_binary_directory
  parse_db["ctest_memcheck"] = parse_ctest_memcheck
  parse_db["ctest_read_custom_files"] = parse_ctest_read_custom_files
  parse_db["ctest_run_script"] = parse_ctest_run_script
  parse_db["ctest_sleep"] = parse_ctest_sleep
  parse_db["ctest_start"] = parse_ctest_start
  parse_db["ctest_submit"] = parse_ctest_submit
  parse_db["ctest_test"] = parse_ctest_test
  parse_db["ctest_update"] = parse_ctest_update
  parse_db["ctest_upload"] = parse_ctest_upload
  parse_db["target_compile_definitions"] = parse_target_compile_definitions
  parse_db["target_compile_features"] = parse_target_compile_features
  parse_db["target_sources"] = parse_target_sources
  parse_db["unset"] = parse_unset
  parse_db["variable_watch"] = parse_variable_watch
  parse_db["add_compile_definitions"] = parse_add_compile_definitions



import collections

from cmakelang import lex
from cmakelang.parse.common import KwargBreaker, NodeType, TreeNode
from cmakelang.parse.argument_nodes import (
    ArgGroupNode, KeywordGroupNode, PositionalGroupNode, PositionalParser)
from cmakelang.parse.util import (
    WHITESPACE_TOKENS,
    get_normalized_kwarg,
    get_tag,
    should_break)

CacheTuple = collections.namedtuple(
    "CacheTuple", ["type", "docstring", "force"])


class SetFnNode(ArgGroupNode):
  def __init__(self):
    super(SetFnNode, self).__init__()
    self.varname = None
    self.value_group = None
    self.cache = None
    self.parent_scope = False


def parse_set(ctx, tokens, breakstack):
  """
  ::

    set(<variable> <value>
        [[CACHE <type> <docstring> [FORCE]] | PARENT_SCOPE])

  :see: https://cmake.org/cmake/help/v3.0/command/set.html?
  """
  tree = SetFnNode()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  kwargs = {
      "CACHE": PositionalParser('2+', flags=["FORCE"])
  }
  flags = ["PARENT_SCOPE"]
  kwarg_breakstack = breakstack + [KwargBreaker(list(kwargs.keys()) + flags)]
  positional_breakstack = breakstack + [KwargBreaker(list(kwargs.keys()))]

  ntokens = len(tokens)
  subtree = PositionalGroupNode.parse(
      ctx, tokens, 1, flags, positional_breakstack)
  assert len(tokens) < ntokens
  tree.children.append(subtree)
  for token in subtree.get_semantic_tokens():
    tree.varname = token
    break

  while tokens:
    # Break if the next token belongs to a parent parser, i.e. if it
    # matches a keyword argument of something higher in the stack, or if
    # it closes a parent group.
    if should_break(tokens[0], breakstack):
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # If it's a comment, then add it at the current depth
    if tokens[0].type in (lex.TokenType.COMMENT,
                          lex.TokenType.BRACKET_COMMENT):
      if not get_tag(tokens[0]) in ("sort", "sortable"):
        child = TreeNode(NodeType.COMMENT)
        tree.children.append(child)
        child.children.append(tokens.pop(0))
        continue

    ntokens = len(tokens)

    word = get_normalized_kwarg(tokens[0])
    if word == "CACHE":
      with ctx.pusharg(tree):
        subtree = KeywordGroupNode.parse(
            ctx, tokens, word, kwargs[word], kwarg_breakstack)
      cache_tokens = subtree.get_semantic_tokens()
      cache_tokens.pop(0)
      cache_args = [None, None, False]
      for idx, token in enumerate(cache_tokens[:2]):
        cache_args[idx] = token.spelling

      if not cache_tokens:
        ctx.lint_ctx.record_lint(
            "E1120", "<variable>", location=tokens[0].get_location())
      elif len(cache_tokens) < 2:
        ctx.lint_ctx.record_lint(
            "E1120", "<value>", location=tokens[0].get_location())
      elif len(cache_tokens) < 3:
        pass
      elif len(cache_tokens) > 4:
        ctx.lint_ctx.record_lint(
            "E1121", location=cache_tokens[2].get_location())
      else:
        if cache_tokens[2].spelling.upper() == "FORCE":
          cache_args[2] = True
        else:
          ctx.lint_ctx.record_lint(
              "E1121", location=cache_tokens[2].get_location())
      tree.cache = CacheTuple(*cache_args)

    elif word == "PARENT_SCOPE":
      with ctx.pusharg(tree):
        subtree = PositionalGroupNode.parse(
            ctx, tokens, '+', ["PARENT_SCOPE"], positional_breakstack)
      tree.parent_scope = True
    else:
      with ctx.pusharg(tree):
        subtree = PositionalGroupNode.parse(
            ctx, tokens, '+', [], kwarg_breakstack)
      for pattern, tags in ctx.config.parse.vartags_:
        if pattern.match(tree.varname.spelling):
          subtree.tags.extend(tags)
      tree.value_group = subtree

    assert len(tokens) < ntokens
    tree.children.append(subtree)
  return tree


def populate_db(parse_db):
  parse_db["set"] = parse_set



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import logging
import os
import unittest
import subprocess
import sys
import tempfile


def get_repo_dir():
  """
  Return the path to the repository root
  """
  thisdir = os.path.dirname(os.path.realpath(__file__))
  parent, _ = os.path.split(thisdir)
  parent, _ = os.path.split(parent)
  return parent


class TestContribution(unittest.TestCase):
  """
  Verify that contributed code meets requirements
  """

  def __init__(self, *args, **kwargs):
    super(TestContribution, self).__init__(*args, **kwargs)
    self.merge_target = None
    self.merge_feature = None
    self.repodir = None

  def setUp(self):
    if os.environ.get("TRAVIS_PULL_REQUEST", "false") == "false":
      self.skipTest("Not a pull request")
      return

    self.repodir = get_repo_dir()

  def test_single_commit(self):
    """
    Ensure that the pull request is a single commit off of the target
    branch. If it is, then the merge base will be the parent of the feature
    commit.
    """
    repodir = get_repo_dir()
    feature_sha1 = os.environ.get("TRAVIS_PULL_REQUEST_SHA")
    feature_parent = subprocess.check_output(
        ["git", "show", "-s", "--pretty=%P", feature_sha1], cwd=repodir
    ).decode("utf-8").strip()
    self.assertEqual(1, len(feature_parent.split()))

    target_branch = os.environ.get("TRAVIS_BRANCH")
    subprocess.check_call(
        ["git", "fetch", "origin", target_branch],
        cwd=self.repodir)
    merge_base = subprocess.check_output(
        ["git", "merge-base", "FETCH_HEAD", feature_sha1],
        cwd=self.repodir).decode("utf-8").strip()
    self.assertEqual(feature_parent, merge_base)

  def test_merge_base_is_not_too_old(self):
    """
    When we merge a pull request we'll get something like this::

           x -----------
          /             \
   ... - A - o - o - o - B

    where x is the signed commit of the change to pull in. If the commit was
    created "a long time ago" (in a git-sense) then this makes the graph
    difficult to read as there will be lots of very long merge-lines. Therefore
    we require that pull requests be "close-enough" to the current HEAD.
    """
    repodir = get_repo_dir()
    feature_sha1 = os.environ.get("TRAVIS_PULL_REQUEST_SHA")
    target_branch = os.environ.get("TRAVIS_BRANCH")
    subprocess.check_call(
        ["git", "fetch", "origin", target_branch],
        cwd=self.repodir)
    merge_base = subprocess.check_output(
        ["git", "merge-base", "FETCH_HEAD", feature_sha1],
        cwd=self.repodir).decode("utf-8").strip()
    distance_str = subprocess.check_output(
        ["git", "rev-list", "--count", "--first-parent",
         "{}..{}".format(merge_base, "FETCH_HEAD")], cwd=repodir
    ).decode("utf-8")
    distance = int(distance_str)
    self.assertLess(distance, 10)

  def test_commit_is_signed(self):
    """
    Ensure that the feature commit is signed
    """
    repodir = get_repo_dir()
    feature_sha1 = os.environ.get("TRAVIS_PULL_REQUEST_SHA")
    signer_encoded = subprocess.check_output(
        ["git", "show", "-s", r'--pretty="%GK"', feature_sha1],
        cwd=repodir).decode("utf-8").strip().strip('"')

    # TODO(josh): deduplicate with validate_database
    # First, get a list of keys already in the local keyring
    known_fingerprints = set()
    proc = subprocess.Popen(
        ["gpg", "--fingerprint", "--with-colons"],
        stdout=subprocess.PIPE)
    for line in proc.stdout:
      parts = line.decode("utf-8").split(":")
      if parts[0] == "fpr":
        fingerprint = parts[-2]
        known_fingerprints.add(fingerprint[-16:])
    proc.wait()

    if signer_encoded not in known_fingerprints:
      # TODO(josh): use SKS pool instead of specific server
      result = subprocess.check_call(
          ["gpg", "--keyserver", "keyserver.ubuntu.com", "--recv-keys",
           signer_encoded])
      self.assertEqual(
          result, 0, msg="Failed to fetch signer key from keyserver")

    with tempfile.NamedTemporaryFile(delete=False) as stderr:
      result = subprocess.call(
          ["git", "verify-commit", feature_sha1],
          cwd=repodir, stderr=stderr)
      stderrpath = stderr.name
    with open(stderrpath, "r") as infile:
      stderrtext = infile.read()
    os.unlink(stderrpath)
    self.assertEqual(
        0, result,
        "git was unable to verify commit {}\n{}"
        .format(feature_sha1, stderrtext))

  def test_commit_signer_has_signed_ca(self):
    """
    Verify that whoever signed the commit has also signed the contributor
    agreement / copyright assignment.
    """
    if sys.version_info < (3, 0, 0):
      self.skipTest("no pgpy on this python version")
    import pgpy  # pylint: disable=import-error

    repodir = get_repo_dir()
    feature_sha1 = os.environ.get("TRAVIS_PULL_REQUEST_SHA")
    signer_encoded = subprocess.check_output(
        ["git", "show", "-s", r'--pretty="%GK"', feature_sha1],
        cwd=repodir).decode("utf-8").strip().strip('"')

    with open(os.path.join(
        repodir, "cmakelang/contrib/signature_db.json")) as infile:
      sigdb = json.load(infile)

    for sigentry in sigdb:
      clearsign = format_signature(sigentry["signature"])
      sig = pgpy.PGPSignature()
      sig.parse(clearsign)
      if sig.signer == signer_encoded:
        return

    self.fail("Could not find signer {} in signature database"
              .format(signer_encoded))


def format_signature(armored_packet):
  """
  Append the common that were stripped out of the signature when stored in the
  database.
  """
  return "\n".join(
      ["-----BEGIN PGP SIGNATURE-----", ""]
      + armored_packet
      + ["-----END PGP SIGNATURE-----"])


if __name__ == '__main__':
  logging.basicConfig(level=logging.DEBUG)
  unittest.main()



# PYTHON_ARGCOMPLETE_OK
"""
Parse cmake help to generate certain input files

TODO(josh): `cmake --help-module-list` to get a list of builtin find-modules
that we can search for more commands.
TODO(josh): `cmake --help-commandplist` to get a list of builtin commands
that we can generate parsers for.
"""

from __future__ import print_function, unicode_literals

import argparse
import io
import logging
import os
import re
import subprocess
import sys

import jinja2

logger = logging.getLogger(__name__)


def get_abspath(relpath):
  return os.path.abspath(os.path.join(os.path.dirname(__file__), relpath))


def sub_callback(match):
  # NOTE(josh): it would be handly to use named groups in regular expressions,
  # but since we use | to join them together into a big regex... the names
  # will collide so we cannot
  return "(?P<{}>.*)".format(re.sub(r"\W", "_", match.group(1)))


GENERIC_LABEL = re.compile(r"<([^>]+)>")


def make_pattern(namestr):
  """
  Look for any generic labels within a cmake property or variable name
  (e.g. `<LANG>` in `<LANG>_CPPLINT`) and convert the pattern string to a
  regular expression pattern (e.g. `.*_CPPLINT`)
  """
  return GENERIC_LABEL.sub(sub_callback, namestr)


def strip_named_groups(pattern):
  return re.sub(r"\(?P<[\w_]>", "(", pattern)


def get_properties(args, jenv):

  proc = subprocess.Popen(
      [args.cmake, "--help-property-list"],
      stdout=subprocess.PIPE)

  with proc.stdout as infile:
    properties = [line.decode("utf-8").strip() for line in infile]
  patterns = [make_pattern(namestr) for namestr in properties]
  proc.wait()

  template = jenv.get_template("properties.jinja.py")
  content = template.render(patterns=patterns)

  if args.outfile == "-":
    args.outfile = os.dup(sys.stdout.fileno())
  with io.open(args.outfile, "w", encoding="utf-8") as outfile:
    outfile.write(content)
    outfile.write("\n")


def get_variables(args, jenv):

  proc = subprocess.Popen(
      [args.cmake, "--help-variable-list"],
      stdout=subprocess.PIPE)

  with proc.stdout as infile:
    variables = [line.decode("utf-8").strip() for line in infile]
  patterns = [make_pattern(namestr) for namestr in variables]
  proc.wait()

  template = jenv.get_template("variables.jinja.py")
  content = template.render(patterns=patterns)

  if args.outfile == "-":
    args.outfile = os.dup(sys.stdout.fileno())
  with io.open(args.outfile, "w", encoding="utf-8") as outfile:
    outfile.write(content)
    outfile.write("\n")


def get_command_list(args):
  """
  Get a list of all the builtin cmake commands (statement names).
  """
  proc = subprocess.Popen(
      [args.cmake, "--help-command-list"],
      stdout=subprocess.PIPE)

  with proc.stdout as infile:
    return [line.decode("utf-8").strip() for line in infile]


def get_command_help(args, command_name):
  """
  Get the help string for a specific command
  """
  proc = subprocess.Popen(
      [args.cmake, "--help-command", command_name],
      stdout=subprocess.PIPE)
  with proc.stdout as infile:
    return infile.read().decode("utf-8")


def get_usages(helpstr):
  """
  Parse the command help string and return a list of command usage strings.
  """
  usage = []
  active = False

  buf = ""
  lineiter = iter(helpstr.split("\n"))
  for line in lineiter:
    if active:
      if not line.strip():
        active = False
        usage.append(buf)
        buf = ""
      else:
        buf += line.rstrip() + "\n"
    elif line.strip() == "::":
      next(lineiter, None)
      active = True

  if buf:
    usage.append(buf)
  return usage


def cmd_get_usages(args, jenv):  # pylint: disable=W0613
  for command_name in get_command_list(args):
    helpstr = get_command_help(args, command_name)
    for usage in get_usages(helpstr):
      print(usage)


def cmd_print_deprecated(args, jenv):  # pylint: disable=W0613
  for command_name in get_command_list(args):
    helpstr = get_command_help(args, command_name)
    if "Deprecated" in helpstr:
      print(command_name)


def setup_argparse(parser):
  """Setup argument parser"""
  parser.add_argument(
      "--cmake", default="cmake",
      help="Cmake binary path or command. Default is `cmake`")
  parser.add_argument(
      "--outfile", default="-",
      help="output file, default is stdout")
  subparsers = parser.add_subparsers(dest="command")
  subparsers.add_parser("properties")
  subparsers.add_parser("variables")
  subparsers.add_parser("usages")
  subparsers.add_parser("deprecated")


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  setup_argparse(parser)
  try:
    import argcomplete
    argcomplete.autocomplete(parser)
  except ImportError:
    pass
  args = parser.parse_args()

  thisdir = os.path.abspath(os.path.dirname(__file__))
  jenv = jinja2.Environment(
      loader=jinja2.FileSystemLoader(thisdir)
  )

  if args.command == "properties":
    get_properties(args, jenv)
  elif args.command == "variables":
    get_variables(args, jenv)
  elif args.command == "usages":
    cmd_get_usages(args, jenv)
  elif args.command == "deprecated":
    cmd_print_deprecated(args, jenv)
  else:
    logger.warning("Unknown command %s", args.command)


if __name__ == "__main__":
  logging.basicConfig(level=logging.INFO)
  main()



import importlib
import os

# Source the common stuff
with open(os.path.join("./conf_common.py")) as infile:
  exec(infile.read())  # pylint: disable=W0122

_module = importlib.import_module("cmakelang")

# Override the project-specific stuff
project = "cmakelang"
docname = project + u'doc'
title = project + ' Documentation'
version = _module.__version__
release = _module.__version__



# pylint: disable=bad-continuation
from __future__ import unicode_literals
import unittest

from cmakelang.command_tests import TestBase


class TestAddExecutableCommand(TestBase):
  """
  Test various examples of add_executable()
  """
  kExpectNumSidecarTests = 6

  def test_sort_arguments(self):
    self.config.format.autosort = True
    self.source_str = """\
add_executable(foobar WIN32 sourcefile_04.cc sourcefile_02.cc sourcefile_03.cc
                            sourcefile_01.cc)
"""
    self.expect_format = """\
add_executable(foobar WIN32 sourcefile_01.cc sourcefile_02.cc sourcefile_03.cc
                            sourcefile_04.cc)
"""


if __name__ == '__main__':
  unittest.main()



import unittest

# pylint: disable=wildcard-import
# pylint: disable=unused-wildcard-import
# pylint: disable=unused-import

from cmakelang.format.invocation_tests import TestInvocations
from cmakelang.format.layout_tests import TestCanonicalLayout
from cmakelang.lex.tests import TestSpecificLexings
from cmakelang.markup_tests import *
from cmakelang.parse.tests import TestCanonicalParse

from cmakelang.command_tests import (
    TestAddCustomCommand,
    TestComment,
    TestConditional,
    TestCustomCommand,
    TestExport,
    TestExternalProject,
    TestForeach,
    TestFile,
    TestInstall,
    TestSetTargetProperties,
    TestSet)

from cmakelang.command_tests.add_executable_tests \
    import TestAddExecutableCommand
from cmakelang.command_tests.add_library_tests \
    import TestAddLibraryCommand
from cmakelang.command_tests.misc_tests \
    import TestMiscFormatting
from cmakelang.contrib.validate_database \
    import TestContributorAgreements
from cmakelang.contrib.validate_pullrequest \
    import TestContribution
from cmakelang.lint.test.expect_tests import (
    ConfigTestCase,
    LintTests)
from cmakelang.lint.test.execution_tests import (
    TestFormatFiles)
from cmakelang.test.version_number_test \
    import TestVersionNumber
from cmakelang.test.command_db_test \
    import TestCommandDatabase
from cmakelang.test.config_include_test \
    import TestConfigInclude

if __name__ == '__main__':
  unittest.main()



from cmakelang.parse.argument_nodes import (
    PositionalParser,
    StandardArgTree,
)
from cmakelang.parse.util import (
    get_nth_semantic_token
)


def parse_build_name(ctx, tokens, breakstack):
  """
  ::

    build_name(variable)

  :see: https://cmake.org/cmake/help/latest/command/build_name.html
  """
  return StandardArgTree.parse(ctx, tokens, 1, {}, [], breakstack)


def parse_exec_program(ctx, tokens, breakstack):
  """
  ::

    exec_program(Executable [directory in which to run]
                 [ARGS <arguments to executable>]
                 [OUTPUT_VARIABLE <var>]
                 [RETURN_VALUE <var>])

  :see: https://cmake.org/cmake/help/latest/command/exec_program.html
  """
  kwargs = {
      "ARGS": PositionalParser("+"),
      "OUTPUT_VARIABLE": PositionalParser(1),
      "RETURN_VALUE": PositionalParser(1)
  }

  return StandardArgTree.parse(ctx, tokens, 2, kwargs, [], breakstack)


def parse_install_files(ctx, tokens, breakstack):
  """
  ::

    install_files(<dir> extension file file ...)
    install_files(<dir> regexp)
    install_files(<dir> FILES file file ...)

  :see: https://cmake.org/cmake/help/latest/command/install_files.html
  """
  second_token = get_nth_semantic_token(tokens, 1)
  third_token = get_nth_semantic_token(tokens, 2)

  if second_token is not None and second_token.spelling.upper() == "FILES":
    return StandardArgTree.parse(ctx, tokens, "3+", {}, ["FILES"], breakstack)

  if third_token is None:
    return StandardArgTree.parse(ctx, tokens, 2, {}, [], breakstack)

  return StandardArgTree.parse(ctx, tokens, "3+", {}, [], breakstack)


def parse_install_programs(ctx, tokens, breakstack):
  """
  ::

    install_programs(<dir> file1 file2 [file3 ...])
    install_programs(<dir> FILES file1 [file2 ...])
    install_programs(<dir> regexp)

  :see: https://cmake.org/cmake/help/latest/command/install_programs.html
  """
  second_token = get_nth_semantic_token(tokens, 1)
  third_token = get_nth_semantic_token(tokens, 2)

  if second_token is not None and second_token.spelling.upper() == "FILES":
    return StandardArgTree.parse(ctx, tokens, "3+", {}, ["FILES"], breakstack)

  if third_token is None:
    return StandardArgTree.parse(ctx, tokens, 2, {}, [], breakstack)

  return StandardArgTree.parse(ctx, tokens, "3+", {}, [], breakstack)


def parse_install_targets(ctx, tokens, breakstack):
  """
  ::

    install_targets(<dir> [RUNTIME_DIRECTORY dir] target target)

  :see: https://cmake.org/cmake/help/latest/command/install_targets.html
  """
  kwargs = {
      "RUNTIME_DIRECTORY": PositionalParser(1)
  }
  return StandardArgTree.parse(ctx, tokens, "2+", kwargs, [], breakstack)


def parse_make_directory(ctx, tokens, breakstack):
  """
  ::

    make_directory(directory)

  :see: https://cmake.org/cmake/help/latest/command/make_directory.html
  """
  return StandardArgTree.parse(ctx, tokens, 1, {}, [], breakstack)


def parse_remove(ctx, tokens, breakstack):
  """
  ::

    remove(VAR VALUE VALUE ...)

  :see: https://cmake.org/cmake/help/latest/command/remove.html
  """
  return StandardArgTree.parse(ctx, tokens, "2+", {}, [], breakstack)


def parse_subdir_depends(ctx, tokens, breakstack):
  """
  ::

    subdir_depends(subdir dep1 dep2 ...)

  :see: https://cmake.org/cmake/help/latest/command/subdir_depends.html
  """
  return StandardArgTree.parse(ctx, tokens, "2+", {}, [], breakstack)


def parse_subdirs(ctx, tokens, breakstack):
  """
  ::

    subdirs(dir1 dir2 ...[EXCLUDE_FROM_ALL exclude_dir1 exclude_dir2 ...]
            [PREORDER] )

  :see: https://cmake.org/cmake/help/latest/command/subdirs.html
  """
  kwargs = {
      "EXCLUDE_FROM_ALL": PositionalParser("+"),
  }
  flags = ["PREORDER"]
  return StandardArgTree.parse(ctx, tokens, "+", kwargs, flags, breakstack)


def parse_use_mangled_mesa(ctx, tokens, breakstack):
  """
  ::

    use_mangled_mesa(PATH_TO_MESA OUTPUT_DIRECTORY)

  :see: https://cmake.org/cmake/help/latest/command/use_mangled_mesa.html
  """
  return StandardArgTree.parse(ctx, tokens, 2, {}, [], breakstack)


def parse_utility_source(ctx, tokens, breakstack):
  """
  ::

    utility_source(cache_entry executable_name
                   path_to_source [file1 file2 ...])

  :see: https://cmake.org/cmake/help/latest/command/utility_source.html
  """
  return StandardArgTree.parse(ctx, tokens, "3+", {}, [], breakstack)


def parse_variable_requires(ctx, tokens, breakstack):
  """
  ::

    variable_requires(TEST_VARIABLE RESULT_VARIABLE
                      REQUIRED_VARIABLE1
                      REQUIRED_VARIABLE2 ...)

  :see: https://cmake.org/cmake/help/latest/command/variable_requires.html
  """
  return StandardArgTree.parse(ctx, tokens, "3+", {}, [], breakstack)


def parse_write_file(ctx, tokens, breakstack):
  """
  ::

    write_file(filename "message to write"... [APPEND])

  :see: https://cmake.org/cmake/help/latest/command/write_file.html
  """
  return StandardArgTree.parse(ctx, tokens, "2+", {}, ["APPEND"], breakstack)


def populate_db(parse_db):
  parse_db["build_name"] = parse_build_name
  parse_db["exec_program"] = parse_exec_program
  parse_db["install_files"] = parse_install_files
  parse_db["install_programs"] = parse_install_programs
  parse_db["install_targets"] = parse_install_targets
  parse_db["make_directory"] = parse_make_directory
  parse_db["remove"] = parse_remove
  parse_db["subdir_depends"] = parse_subdir_depends
  parse_db["subdirs"] = parse_subdirs
  parse_db["use_mangled_mesa"] = parse_use_mangled_mesa
  parse_db["utility_source"] = parse_utility_source
  parse_db["variable_requires"] = parse_variable_requires
  parse_db["write_file"] = parse_write_file



"""
Generate the copyright assignment document from the template, and optionally
sign it with gpg.
"""

import argparse
import collections
import io
import json
import os
import subprocess
import sys
import tempfile


def get_signature_packet(infile):
  """
  Return the ascii-armored signature packet between the lines:
  -----BEGIN PGP SIGNATURE-----

  <packet-content>
  -----END PGP SIGNATURE-----
  """

  lineiter = iter(infile)
  for line in lineiter:
    if "BEGIN PGP SIGNATURE" in line:
      break

  packetlines = []
  next(lineiter)
  for line in lineiter:
    if "END PGP SIGNATURE" in line:
      return packetlines
    packetlines.append(line.strip())

  raise ValueError("no PGP BEGIN/END tags in file content!")


def generate_document(args):
  thisdir = os.path.dirname(os.path.realpath(__file__))
  with io.open(os.path.join(thisdir, "individual_ca.txt"),
               encoding="utf-8") as infile:
    output_text = infile.read()
  output_text = output_text.replace(r"{{signer_name}}", args.name)
  output_text = output_text.replace(r"{{signer_email}}", args.email)

  with io.open(args.output_text, "w", encoding="utf-8") as outfile:
    outfile.write(output_text)

  if args.only_template:
    print(
        "The contributor agreement has been written to {}, please sign it with "
        "GPG and include the signature in your first contribution"
        .format(args.output_text))
    return

  argv = ["gpg", "--output", args.output_sig]
  if args.signkey is not None:
    argv.extend(["--default-key", args.signkey])
  argv.extend(["--armor", "--detach-sign", args.output_text])
  subprocess.call(argv)
  print(
      "The contributor agreement has been written to {}, and the signature "
      "has been written to {}"
      .format(args.output_text, args.output_sig))


def setup_parser(parser):
  """
  Configure command line parser
  """
  default_name = None
  try:
    default_name = subprocess.check_output(
        ["git", "config", "user.name"]).decode("utf-8").strip()
  except subprocess.CalledProcessError:
    pass

  default_email = None
  try:
    default_email = subprocess.check_output(
        ["git", "config", "user.email"]).decode("utf-8").strip()
  except subprocess.CalledProcessError:
    pass

  default_signkey = None
  try:
    default_signkey = subprocess.check_output(
        ["git", "config", "user.signingkey"]).decode("utf-8").strip()
  except subprocess.CalledProcessError:
    pass

  parser.add_argument(
      "-n", "--name",
      default=default_name,
      required=(default_name is None),
      help=("Your name (to fill in the agreement template). Default={}"
            .format(default_name)))

  parser.add_argument(
      "-e", "--email",
      default=default_email,
      required=(default_email is None),
      help=("Your name (to fill in the agreement template). Default={}"
            .format(default_email)))

  parser.add_argument(
      "-k", "--signkey",
      default=default_signkey,
      help=("Which signing key to use if you don't want to use your gpg "
            "default. Default={}".format(default_signkey)))

  default_outpath = os.path.join(tempfile.gettempdir(), "cmake-format-ca.txt")
  default_sigout = os.path.join(tempfile.gettempdir(), "cmake-format-ca.sig")
  parser.add_argument(
      "-o", "--output-text", default=default_outpath,
      help=r"Where to write out the agreement text. Default is"
           r"/{tempdir}/cmake-format-ca.txt")
  parser.add_argument(
      "-s", "--output-sig", default=default_sigout,
      help=r"Where to write out the signature. Default is"
           r"/{tempdir}/cmake-format-ca.sig")

  parser.add_argument(
      "-t", "--only-template", action="store_true",
      help="Don't attempt to sign the document with gpg, just fill in the"
           " template and write to a file.")


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  setup_parser(parser)
  args = parser.parse_args()
  generate_document(args)
  if args.only_template:
    return 0

  with io.open(args.output_sig, "r", encoding="utf-8") as infile:
    packetlines = get_signature_packet(infile)

  dataitem = collections.OrderedDict([
      ("template", "individual"),
      ("version", "1.0"),
      ("name", args.name),
      ("email", args.email),
      ("signature", packetlines)
  ])

  thisdir = os.path.dirname(os.path.realpath(__file__))
  database_path = os.path.join(thisdir, "signature_db.json")
  with io.open(database_path, "r", encoding="utf-8") as infile:
    database = json.load(infile, object_pairs_hook=collections.OrderedDict)
  database.append(dataitem)
  with io.open(database_path, "w", encoding="utf-8") as outfile:
    json.dump(database, outfile, indent=2)
  return 0


if __name__ == "__main__":
  sys.exit(main())



from __future__ import unicode_literals

import collections
import contextlib
import inspect
import logging
import pprint
import sys
import textwrap

import six

logger = logging.getLogger(__name__)


def serialize(obj, with_help=False, with_defaults=True):
  """
  Return a serializable representation of the object. If the object has an
  `as_dict` method, then it will call and return the output of that method.
  Otherwise return the object itself.
  """
  if hasattr(obj, 'as_odict'):
    fun = getattr(obj, 'as_odict')
    if callable(fun):
      return fun(with_help=with_help, with_defaults=with_defaults)
  if hasattr(obj, 'as_dict'):
    fun = getattr(obj, 'as_dict')
    if callable(fun):
      return fun(with_help=with_help, with_defaults=with_defaults)

  return obj


def parse_bool(string):
  if string.lower() in ('y', 'yes', 't', 'true', '1', 'yup', 'yeah', 'yada'):
    return True
  if string.lower() in ('n', 'no', 'f', 'false', '0', 'nope', 'nah', 'nada'):
    return False

  logger.warning("Ambiguous truthiness of string '%s' evalutes to 'FALSE'",
                 string)
  return False


def get_default(value, default):
  """
  return ``value`` if it is not None, else default
  """
  if value is None:
    return default

  return value


class ExecGlobal(dict):
  """
  We pass this in as the "global" variable when parsing a config. It is
  composed of a nested dictionary and provides methods for mangaging
  the current  "stack" into that nested dictionary
  """

  def __init__(self, filepath):
    super(ExecGlobal, self).__init__()

    # The current stack of push/pop config sections
    self._dict_stack = []

    # This is what we provide for __file__
    self.filepath = filepath

  @contextlib.contextmanager
  def section(self, key):
    self.push_section(key)
    yield None
    self.pop_section()

  def push_section(self, key):
    key = key.replace("-", "_")
    if self._dict_stack:
      stacktop = self._dict_stack[-1]
    else:
      stacktop = self

    if key not in stacktop:
      stacktop[key] = {}
    newtop = stacktop[key]
    self._dict_stack.append(newtop)

  def pop_section(self):
    self._dict_stack.pop(-1)

  def __getitem__(self, key):
    if key in ("_dict_stack", "__name__"):
      # Disallow config script from accessing internal members
      logger.warning("Config illegal attempt to access %s", key)
      return None

    if key == "__file__":
      # Allow config script to access __file__
      return self.filepath

    if hasattr(self, key):
      # Allow config script to access section(), push_section(), pop_section()
      # directly
      return getattr(self, key)

    # Check if any nested scopes contain this name, and return it if they do.
    # this means that sections are "scoped".
    for dictitem in reversed(self._dict_stack):
      if key in dictitem:
        return dictitem[key]

    # Otherwise act like a dictionary
    return super(ExecGlobal, self).__getitem__(key)

  def __setitem__(self, key, value):
    if self._dict_stack:
      self._dict_stack[-1][key] = value
    else:
      super(ExecGlobal, self).__setitem__(key, value)


def _emulate_setname(subclass):
  # pylint: disable=protected-access
  for key, value in subclass.__dict__.items():
    if key.startswith("_"):
      continue

    if hasattr(value, "__set_name__"):
      getattr(value, "__set_name__")(subclass, key)


class ConfigMeta(type):
  """Metaclass that calls __set_name__ on all subclass descriptors for pythons
  prior to 3.6
  """

  def __new__(mcs, name, bases, dct):
    subclass = type.__new__(mcs, name, bases, dct)
    if sys.version_info < (3, 6, 0):
      _emulate_setname(subclass)
    return subclass

  def __setattr__(cls, name, value):
    if hasattr(value, "__set_name__"):
      getattr(value, "__set_name__")(cls, name)


class Descriptor(object):
  # TODO(josh): consider storing a global construction order among
  # descriptors. This would allow us to iterate over config object descriptors
  # in deterministic (construction) order even in pythons where
  # __dict__.items() does not return objects in declaration order.
  pass


class SubtreeDescriptor(Descriptor):
  """Implements the descriptor interface for nested configuration
     objects."""

  def __init__(self, subtree_class):
    super(SubtreeDescriptor, self).__init__()
    self.name = "<unnamed>"
    self.subtree_class = subtree_class

  def get(self, obj):
    if not hasattr(obj, "_" + self.name):
      # Unlike for config variables, for subtree objects we always instanciate
      # the shadow member.
      setattr(obj, "_" + self.name, self.subtree_class())

    return getattr(obj, "_" + self.name)

  def __get__(self, obj, _objtype):
    return self.get(obj)

  def __set__(self, obj, value):
    raise NotImplementedError(
        "Assignment to subtree objects is not supported")

  def __set_name__(self, owner, name):
    # pylint: disable=protected-access
    owner._field_registry.append(self)
    if sys.version_info < (3, 0, 0):
      name = name.decode("utf-8")
    self.name = name

  def consume_value(self, obj, value_dict):
    """Convenience method to consume keyword arguments directly from the
    descriptor interface."""
    if not isinstance(value_dict, dict):
      raise ValueError(
          "value_dict for {}.{} must be a dictionary, not a {}"
          .format(type(obj).__name__, self.name, type(value_dict).__name__))
    self.get(obj).consume_known(value_dict)
    warn_unused(value_dict)

  def legacy_shim_consume(self, obj, kwargs):
    """Consume config variable assignments from the root of the config
    tree. This is the legacy config style and will likely be deprecated
    soon."""
    self.get(obj).legacy_consume(kwargs)

  def add_to_argparser(self, argparser):
    self.subtree_class.add_to_argparser(argparser)


if sys.version_info >= (3, 0, 0):
  VALUE_TYPES = (str, int, float)
else:
  VALUE_TYPES = (str, unicode, int, float)


class FieldDescriptor(Descriptor):
  """Implements the descriptor interface to store metadata (default value,
  docstring, and choices) for each configuration variable.
  """

  def __init__(self, default_value=None, helptext=None, choices=None):
    super(FieldDescriptor, self).__init__()
    self.default_value = default_value
    self.helptext = helptext
    self.choices = choices
    self.name = "<unnamed>"

  def __get__(self, obj, objtype):
    return getattr(obj, "_" + self.name, self.default_value)

  def __set__(self, obj, value):
    setattr(obj, "_" + self.name, value)

  def unset(self, obj):
    if self.has_override(obj):
      delattr(obj, "_" + self.name)

  def __set_name__(self, owner, name):
    # pylint: disable=protected-access
    owner._field_registry.append(self)
    if sys.version_info < (3, 0, 0):
      name = name.decode("utf-8")
    self.name = name

  def consume_value(self, obj, value):
    """Convenience method to consume values directly from the descriptor
    interface."""
    self.__set__(obj, value)

  def has_override(self, obj):
    """Return true if `obj` has an override for this configuration variable."""
    return hasattr(obj, "_" + self.name)

  def add_to_argparse(self, optgroup):
    """Add the config variable as an argument to the command line parser."""
    if self.name == 'additional_commands':
      return

    if isinstance(self.default_value, bool):
      # NOTE(josh): argparse store_true isn't what we want here because we
      # want to distinguish between "not specified" = "default" and
      # "specified"
      optgroup.add_argument(
          '--' + self.name.replace('_', '-'), nargs='?',
          default=None, const=(not self.default_value),
          type=parse_bool, help=self.helptext)
    elif isinstance(self.default_value, VALUE_TYPES):
      optgroup.add_argument(
          '--' + self.name.replace('_', '-'),
          type=type(self.default_value),
          help=self.helptext,
          choices=self.choices)
    elif self.default_value is None:
      # If the value is None then we can't really tell what it's supposed to
      # be. I guess let's assume string in this case.
      optgroup.add_argument(
          '--' + self.name.replace('_', '-'),
          help=self.helptext,
          choices=self.choices)
    # NOTE(josh): argparse behavior is that if the flag is not specified on
    # the command line the value will be None, whereas if it's specified with
    # no arguments then the value will be an empty list. This exactly what we
    # want since we can ignore `None` values.
    elif isinstance(self.default_value, (list, tuple)):
      typearg = None
      if self.default_value:
        typearg = type(self.default_value[0])
      optgroup.add_argument(
          '--' + self.name.replace('_', '-'),
          nargs='*', type=typearg, help=self.helptext)


def serialize_docstring(docstring):
  if "\n" in docstring:
    return docstring.split("\n")
  return docstring


def warn_unused(kwargs):
  unused = []
  for keyword, value in kwargs.items():
    if keyword.startswith("_"):
      continue
    if inspect.ismodule(value):
      continue
    unused.append(keyword)

  if unused:
    logger.warning(
        "The following configuration options were ignored:\n  %s",
        "\n  ".join(unused))


class ConfigObject(six.with_metaclass(ConfigMeta)):
  """
  Base class for encapsulationg options/parameters
  """

  _field_registry = []

  def _update_derived(self):
    """subclass hook to update any derived values after a change."""

  def __init__(self, **kwargs):
    # Ensure that the most derived class has a field registry
    assert hasattr(self.__class__, "_field_registry"), (
        "Derived class {} is missing _field_registry member".format(
            self.__class__.__name__))

    self.consume_known(kwargs)
    self.legacy_consume(kwargs)
    warn_unused(kwargs)

  def consume_known(self, kwargs):
    """Consume known configuration values from a dictionary. Removes the
       values from the dictionary.
    """
    for descr in self._field_registry:
      if descr.name in kwargs:
        descr.consume_value(self, kwargs.pop(descr.name))
    self._update_derived()

  def legacy_consume(self, kwargs):
    """Consume arguments from the root of the configuration dictionary.
    """
    self.consume_known(kwargs)
    for descr in self._field_registry:
      if isinstance(descr, SubtreeDescriptor):
        descr.legacy_shim_consume(self, kwargs)

  @classmethod
  def get_field_names(cls):
    return [descr.name for descr in cls._field_registry]

  def _as_dict(self, dtype, with_help=False, with_defaults=True):
    """
    Return a dictionary mapping field names to their values only for fields
    specified in the constructor
    """
    out = dtype()
    for descr in self._field_registry:
      value = descr.__get__(self, type(self))
      if isinstance(descr, SubtreeDescriptor):
        if not (with_defaults or value.has_override()):
          continue
        if with_help:
          out["_help_" + descr.name] = serialize_docstring(value.__doc__)
        # pylint: disable=protected-access
        out[descr.name] = value._as_dict(dtype, with_help, with_defaults)
      elif isinstance(descr, FieldDescriptor):
        if not (with_defaults or descr.has_override(self)):
          continue
        helptext = descr.helptext
        if helptext and with_help:
          out["_help_" + descr.name] = textwrap.wrap(helptext, 60)
        out[descr.name] = serialize(value)
      else:
        raise RuntimeError("Field registry contains unknown descriptor")
    return out

  def as_dict(self, with_help=False, with_defaults=True):
    return self._as_dict(dict, with_help, with_defaults)

  def as_odict(self, with_help=False, with_defaults=True):
    return self._as_dict(collections.OrderedDict, with_help, with_defaults)

  def has_override(self):
    for descr in self._field_registry:
      if isinstance(descr, SubtreeDescriptor):
        if descr.get(self).has_override():
          return True
      elif isinstance(descr, FieldDescriptor):
        if descr.has_override(self):
          return True
      else:
        raise RuntimeError("Field registry contains unknown descriptor")
    return False

  def dump(self, outfile, depth=0, with_help=True, with_defaults=True):
    indent = "  " * depth
    linewidth = 80 - len(indent)
    comment_linewidth = 80 - len(indent) - len("# ")

    ppr = pprint.PrettyPrinter(indent=2, width=linewidth)
    for descr in self._field_registry:
      fieldname = descr.name
      value = descr.__get__(self, type(self))

      if isinstance(descr, SubtreeDescriptor):
        if not (with_defaults or value.has_override()):
          continue

        if value.__doc__ and with_help:
          section_doc_lines = value.__doc__.split("\n")
          ruler_len = max(len(line) for line in section_doc_lines)
          outfile.write("# {}\n".format("-" * ruler_len))
          for line in section_doc_lines:
            outfile.write(indent)
            outfile.write('# {}\n'.format(line.rstrip()))
          outfile.write("# {}\n".format("-" * ruler_len))

        outfile.write(indent)
        outfile.write("with section(\"{}\"):\n".format(fieldname))
        value.dump(outfile, depth + 1, with_help, with_defaults)
        outfile.write("\n")
      elif isinstance(descr, FieldDescriptor):
        if not (with_defaults or descr.has_override(self)):
          continue

        helptext = descr.helptext
        if helptext and with_help:
          outfile.write("\n")
          for line in textwrap.wrap(helptext, comment_linewidth):
            outfile.write(indent)
            outfile.write('# {}\n'.format(line.rstrip()))

        pretty = "{} = {}".format(fieldname, ppr.pformat(value))
        for line in pretty.split("\n"):
          outfile.write(indent)
          outfile.write(line)
          outfile.write("\n")
      else:
        raise RuntimeError("Field registry contains unknown descriptor")

  @classmethod
  def add_to_argparser(cls, argparser):
    if "\n" in cls.__doc__:
      title, description = cls.__doc__.split("\n", 1)
    else:
      title = cls.__doc__
      description = None

    optgroup = argparser.add_argument_group(
        title=title, description=description)

    for descr in cls._field_registry:
      if isinstance(descr, SubtreeDescriptor):
        descr.add_to_argparser(argparser)
      elif isinstance(descr, FieldDescriptor):
        descr.add_to_argparse(optgroup)

  def clone(self):
    """
    Return a copy of self.
    """
    return self.__class__(**self.as_dict())

  def validate(self):
    # TODO(josh): recurse on sub objects
    # TODO(josh): for any spec item that has choices, make sure the value is
    # in the choices
    # TODO(josh): make sure that the current value is the same type as the
    # default value (unless the default is None)
    return True
